from datetime import datetime
from session import SshSession
from utils.decorator import root_access_required

class SyslogService:
    """
    Class for controlling the Syslog Server.

        Args:
            ssh_session (SshSession): SSH session object.
            sudo_paswd (str): Password for sudo command.

        Attributes:
            result_fname (str): File name of the result log file.
            _sudo_pwd (str): Password for sudo command.
            _ssh (SshSession): SSH session object.
    """

    def __init__(self, ssh_session: SshSession, sudo_paswd: str):
        self._ssh = ssh_session
        # TODO: clean log file
        self.result_fname = None
        self._sudo_pwd = sudo_paswd

    def _check_conneted(self):
        """
        Checks if SSH session is connected.

            Raises:
                ValueError: If SSH session is not connected.
        """
        if not self._ssh.is_connected():
            raise ValueError('SSH session not connected')

    def check_env(self):
        """
        Checks if the environment is ready for syslog-ng container to be started.

        Checks if the SSH session is connected and if the docker image balabit/syslog-ng exists in the environment.
            
            Raises:
                ValueError: If SSH session is not connected or docker image balabit/syslog-ng does not exist in the environment. 
        """
        self._check_conneted()
        output = self._ssh.command('docker images')
        if 'balabit/syslog-ng' not in output:
            raise ValueError('No docker image balabit/syslog-ng')

    @root_access_required
    def start(self):
        """
        Starts syslog-ng container on the remote machine.

        Creates a directory and a messages file, logs in as root, and runs a docker command to start the syslog-ng container on the remote machine. 

            Raises: 
                ValueError: If SSH session is not connected. 
        """
        self._check_conneted()
        _now = datetime.now().isoformat()
        self._container_name = 'syslog'
        self.dir_name = 'syslog'
        self._ssh.command(f'mkdir //{self.dir_name}')
        self._ssh.command(f'echo "" > //{self.dir_name}/messages')
        cmd = f'docker run -it --rm -d --name={self._container_name}'\
            f' -p 514:514/udp -v //{self.dir_name}:/var/log/'\
            ' -p 601:601 balabit/syslog-ng:latest'

        self._ssh.command(cmd)

    def stop(self):
        """
        Stop the container.
        """
        self._check_conneted()

        cmd = f'docker stop {self._container_name}'
        self._ssh.command(cmd)
        self._ssh.command('')  # wait prompt

    def read_message(self):
        """
        Reads the messages from the container.

            Returns: 
                str: The messages from the container.
        """
        self._check_conneted()
        cmd = f'docker exec -it {self._container_name} bash -c ' \
            '"cat /var/log/messages"'
        return self._ssh.command(cmd)




from datetime import datetime
from session import SshSession

class TsharkService:
    """
    Class for providing control over tshark using an SSH session.

        Args:
            ssh_session (SshSession): An SSH session object.
            sudo_paswd (str): The sudo password for the SSH session.
        
        Attributes
            _ssh (SshSession): An SSH session object.
            result_fname (str, optional): The name of the result file.  
            _is_capturing (bool):  A boolean indicating whether tshark is currently capturing or not.  
    """

    def __init__(self, ssh_session: SshSession, sudo_paswd: str):
        self._ssh = ssh_session
        # TODO: clean log file
        self.result_fname = None
        self._is_capturing = False

    def _check_conneted(self):
        """
        Check if the environment is valid.

            Raises:
                ValueError: If the SSH session is not connected or if no docker image cincan/tshark is found.
        """
        if not self._ssh.is_connected():
            raise ValueError('SSH session not connected')

    def check_env(self):
        """
        Check if the environment is ready for running TShark.

            Raises:
                ValueError: If no docker image cincan/tshark is present.
        """
        self._check_conneted()
        output = self._ssh.command('docker images')
        if 'cincan/tshark' not in output:
            raise ValueError('No docker image cincan/tshark')

    def _create_command(self,
                        capture_number=None,
                        interface='any',
                        capture_filter=None,
                        display_filter=None,
                        verbose=False,
                        additional_options=None,
                        option=None):
        """
        Create a command for executing tshark in a docker container.

            Args:
                capture_number (int, optional): Number of packets to capture. Defaults to None.
                interface (str, optional): Interface to capture on. Defaults to 'any'.
                capture_filter (str, optional): Capture filter expression. Defaults to None.
                display_filter (str, optional): Display filter expression. Defaults to None. 
                verbose (bool, optional): add output of packet tree (Packet Details). Defaults to False. 
                option (str, optional): Specifying protocol-specific preferences. Defaults to None. 
            Returns: 
                str: Command for executing tshark in a docker container.
        """
        now = datetime.now().isoformat()
        self.result_fname = f'tshark_result_{now}.log'

        cmd = 'docker run -it --rm --cap-add NET_ADMIN --network host ' \
              'cincan/tshark -n '
        cmd += '' if capture_number is None else f'-c {capture_number} '
        cmd += '' if interface is None else f'-i "{interface}" '
        cmd += '' if capture_filter is None else f'-f "{capture_filter}" '
        cmd += '' if display_filter is None else f'-Y "{display_filter}" '
        cmd += '' if verbose is False else '-V '
        cmd += '' if additional_options is None else f'{additional_options} '
        cmd += '' if option is None else f"-o '{option}' "
        cmd += f'> {self.result_fname}'
        return cmd

    def start(self,
              capture_number=None,
              interface: str = 'any',
              capture_filter=None,
              display_filter=None,
              verbose=False,
              additional_options=None,
              option=None):
        """
        Start capturing packets on the remote host.

            Args:
                capture_number (int, optional): Number of packets to capture. Defaults to None.
                interface (str): The network interface to capture packets from. Defaults to 'any'.
                capture_filter (str): A capture filter to apply when capturing packets. Defaults to None. 
                display_filter (str): A display filter to apply when displaying packets. Defaults to None.
                verbose (bool, optional): add output of packet tree (Packet Details). Defaults to False. 
            Returns: 
                bool: True if capturing started successfully, False otherwise. 
        """
        self._check_conneted()
        if self._is_capturing:
            return False
        self._is_capturing = True

        cmd = self._create_command(capture_number=capture_number,
                                   interface=interface,
                                   capture_filter=capture_filter,
                                   display_filter=display_filter,
                                   verbose=verbose,
                                   additional_options=additional_options,
                                   option=option)
        self._ssh.command(cmd)
        return True

    def stop(self):
        """
        Stop the current capture process.

            Returns:
                bool: True if successfully stopped, False otherwise. 
        """
        self._check_conneted()
        if not self._is_capturing:
            return False
        self._ssh.sendcontrol('c')
        self._ssh.command('')  # wait prompt
        self._is_capturing = False
        return True

    def capture_by_number(self,
                          number: str,
                          timeout: str,
                          interface: str = 'any',
                          capture_filter: str = None,
                          display_filter: str = None,
                          additional_options:str = None):
        """
        Start a capture process with given parameters.

            Args: 
                number (str): Number of packets to be captured. 
                timeout (str): Timeout for the capture process in seconds. 
                interface (str): Interface to capture from. Default is 'any'. 
                capture_filter (str): Capture filter string for the capture process. Default is None. 
                display_filter (str): Display filter string for the captured packets. Default is None.  

            Returns: 
                bool: True if successfully started, False otherwise.  
        """
        self._check_conneted()
        if self._is_capturing:
            return False
        self._is_capturing = True

        cmd = self._create_command(capture_number=number,
                                   interface=interface,
                                   capture_filter=capture_filter,
                                   display_filter=display_filter,
                                   additional_options=additional_options)
        r = self._ssh.command(cmd, timeout=timeout)
        
        return True

    def retrieve(self):
        """
        Retrieve the captured packets from remote device as a string.

            Returns: 
                str: Captured packets as a string or None if failed to retrieve them.
        """
        self._check_conneted()
        return self._ssh.command(f'cat {self.result_fname}')



import time


@then(u'mac entry should be recorded on "{device}"')
def step_impl(context, device):
    # make sure the mac table already update
    for _ in range(10):
        flag = False
        mac_table = context.dut[device].main().show_mac_address_table()
        
        for entry in mac_table:
            entry["mac"] = entry["mac"].replace("-", ":").lower()
            if context.mac_info["mac"][0] in entry["mac"]:
                flag = True
                break
        if flag is True: break
        time.sleep(1)

    # confirm the mac table info is correct or not
    for mac in context.mac_info["mac"]:
        flag = False
        for entry in mac_table:
            entry["mac"] = entry["mac"].replace("-", ":").lower()

            if mac in entry["mac"] and \
                context.mac_info["vlan"] in entry["vlan"] and \
                context.mac_info["port"] in entry["port"]:
                flag = True
                break

        assert flag, \
            f'expect-> [mac: {context.mac_info["mac"]}, \
                port: {context.mac_info["port"]}\n but {mac_table}'


@then(u'mac entry should be cleared within the aging time on "{device}"')
def step_impl(context, device):
    ret = context.dut[device].main()._s.command_expect("show mac-address-table")

    # mac-address-table aging time: 5~300 seconds
    aging_time = context.table[0]["aging_time"]
    context.dut[device].go_config().command(f'mac-address-table aging-time {aging_time}')
    time.sleep(int(aging_time))
    
    flag = True
    for mac in context.mac_info["mac"]:
        if mac in ret["data"]:
            flag = False

    assert flag, \
        f'mac table could not be cleared within the aging time: {ret["data"]}'

    
    


from mdc.router.cli.rp_base.ng_router import config, config_if_wan, config_if_lan
from mdc.router.cli.rp_base.ng_router.cli import PROMPT_CONFIG

class ConfigOspf(config.Config):
    """
    Class for configuring Ospf.

        Args:
            session (object): The session object.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Attributes:
            _s (Session): A Session object used to communicate with the device.
            _args (list): A list of arguments passed to the function.
            _kwargs (dict): A dictionary of keyword arguments passed to the function.
    """

    def __init__(self, session: object, *args, **kwargs):
        super().__init__(session)
        self._session = session
        self._args = args
        self._kwargs = kwargs

    def _set_default_prompts(self):
        """
        Set the default prompts.
        """
        self._s.set_default_prompts([PROMPT_CONFIG])

    # Reserve flexibility for function override
    def _getin(self):
        """
        Get into the router configuration mode.
        """
        # default apply the first element as parameter
        routerId = self._args[0] if len(self._args) != 0 else None
        if routerId is None:
            raise Exception(f"Input parameter is empty!")
        self._s.command_expect(f'router ospf {routerId}')

    def getin(self):
        """
        Set the default prompts and get into the interface wan.

            Returns: 
                self: The instance of the object. 
        """
        self._set_default_prompts()
        self._getin()
        return self

    def set_ospf_area(self, area_id: str):
        if area_id:
            self._s.command_expect(f'area {area_id}')
        return self
    
    def set_ospf_redistribute(self, redistribute: str = None):
        if redistribute:
            self._s.command_expect(f'redistribute {redistribute}')
        return self

    def set_ospf_intf(self, 
                      intf: str, 
                      area_id: str, 
                      priority: int = 1,
                      hello_interval: int = 10, 
                      dead_interval: int = 40):
        if intf.lower() == "wan":
            config_wan = config_if_wan.ConfigIfWan(self._session)
            config_wan.getin()
        elif intf.lower() == "lan":
            config_lan = config_if_lan.ConfigIfLan(self._session)
            config_lan.getin()
        else:
            raise ValueError(f"input invalid {intf}")
        
        if area_id:
            self._s.command_expect(f'ip ospf area {area_id}')
        if priority:
            self._s.command_expect(f'ip ospf priority {priority}')
        if hello_interval:
            self._s.command_expect(f'ip ospf hello-interval {hello_interval}')
        if dead_interval:
            self._s.command_expect(f'ip ospf dead-interval {dead_interval}')


from scapy.layers.inet import IP, UDP, ICMP, TCP, ARP, Ether
from scapy.all import *

ripentries = {ripentries}
# 建立RIP Request封包
if "{rip_type}" == "request":
    packet = Ether(src="{smac}", dst="{dmac}") / IP(src="{sip}", dst="224.0.0.9") / UDP(dport=520, sport=520) / RIP(cmd=1, version=2,) / RIPEntry(AF=2, addr="0.0.0.0", mask="0.0.0.0", nextHop="0.0.0.0", metric=16)
# 建立RIP Response封包
elif "{rip_type}" == "response":
    ripentry = RIPEntry(AF=2, addr=ripentries[0]["addr"], mask=ripentries[0]["mask"], nextHop=ripentries[0]["nexthop"], metric=int(ripentries[0]["metric"]))
    for i in range(1, len(ripentries)):
        ripentry /= RIPEntry(AF=2, addr=ripentries[i]["addr"], mask=ripentries[i]["mask"], nextHop=ripentries[i]["nexthop"], metric=int(ripentries[i]["metric"]))
    packet = Ether(src="{smac}", dst="{dmac}") / IP(src="{sip}", dst="224.0.0.9") / UDP(dport=520, sport=520) / RIP(cmd=2,version=2) / ripentry

# 發送RIP Request封包
sendp(packet, iface="{iface}")



from mdc.router.cli.rp_base.lg_router.tn5000 import config
from mdc.router.cli.rp_base.lg_router.tn5000.cli import PROMPT_CONFIG

class ConfigL2filter(config.Config):
    """
    Class for configuring L2 filter.

        Args:
            session (object): A valid session object.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Attributes:
            _s (Session): A Session object used to communicate with the device.
            _args (list): A list of arguments passed to the function.
            _kwargs (dict): A dictionary of keyword arguments passed to the function.
    """

    def __init__(self, session, *args, **kwargs):
        super().__init__(session)
        self._args = args
        self._kwargs = kwargs

    def _set_default_prompts(self):
        self._s.set_default_prompts([PROMPT_CONFIG])

    # Reserve flexibility for function override
    def _getin(self):
        # default apply the first element as parameter
        policy_index = self._args[0] if len(self._args) != 0 else None
        if policy_index is None:
            raise Exception(f"Input parameter is empty!")
        self._s.command_expect(f'l2-filter {policy_index}')

    def getin(self):
        self._set_default_prompts()
        self._getin()
        return self

    def action(self, action):
        action_list = ["accept", "drop"]

        if action in action_list:
            self._s.command_expect(f'action {action}')
        else:
            raise ValueError('action can be either accept or drop')

        return self

    def protocol(self, protocol="all"):
        self._s.command_expect(f'protocol {protocol}')

        return self

    def ether_type(self, ether_type):
        self._s.command_expect(f'ether-type {ether_type}')
        return self

    def src_mac(self, src_mac):
        self._s.command_expect(f'src-mac {src_mac}')
        return self

    def dst_mac(self, dst_mac):
        self._s.command_expect(f'dst-mac {dst_mac}')
        return self

    def interface(self, if_from="all", if_to="all"):
        self._s.command_expect(f'interface {if_from} {if_to}')
        return self

    def exit(self):
        ret = self._s.command_expect('exit',
                                        prompts=[PROMPT_CONFIG],
                                        timeout=2)
        return self




from session import SshSession
from utils.decorator import root_access_required

class TftpService:
    """
    A class to manage tftp server.

        Attributes:
            _ssh (SshSession): An instance of SshSession.
            _sudo_pwd (str): Password for sudo user.
    """

    def __init__(self, ssh_session: SshSession, sudo_paswd: str):
        self._ssh = ssh_session
        self._sudo_pwd = sudo_paswd

    @root_access_required
    def check_env(self):
        """
        Check the environment of tftp server.
        Prints 'TftpServer: check_env' and logs in as root with given password to show status of tftpd-hpa service and log out from root afterwards. Asserts that the service is active. 
        """
        if not self._ssh.is_connected():
            raise ValueError('SSH session not connected')
        ret = self._ssh.command(
            'systemctl status tftpd-hpa.service | grep Active', timeout=5)
        # print(f'ret: {ret}')  # debug
        assert 'active' in ret, f'tftp server is not actived: {ret}'




from mdc.router.cli.rp_base.ng_router.base import Base
from mdc.router.cli.rp_base.ng_router.cli import PROMPT_MAIN

class edr8010_v3_0(Base):

    class Main(Base.Main):

        def set_backdoor(self,
                         account='ieisecureedr_moxasupport',
                         password='83dafc480d466ee0'):
            cmd_args = [
                ('moxasupport 1234', [PROMPT_MAIN]),
                ('exit', [r'login']),
                (account, [r'Password']),
                (password, [r'#']),
            ]
            for args in cmd_args:
                r = self._s.command_expect(args[0], prompts=args[1])
                if not r['matched']:
                    print(f"return = {r}, command = {args[0]}")
                    raise Exception("failed to setup backdoor")
            return self

        def clear_ssh_telnet_limit(self):
            '''
            This step may require modification for future usage.
            Due to the DUT's limit on the number of SSH and Telnet sessions, 
            we need to clear the limitation before logging in and out multiple times.
            '''
            self.set_backdoor()
            self._s.command_expect('iptables -F PortScan', prompts=['#'])
            self._s.command_expect('exit', prompts=['login'])
            # disable ctrl-t mode for serial control
            self.login()



from mdc.router.cli.rp_base.ng_router import config
from mdc.router.cli.rp_base.ng_router.cli import PROMPT_CONFIG

class ConfigRip(config.Config):
    """
    Class for configuring Rip.

        Args:
            session (object): The session object.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Attributes:
            _s (Session): A Session object used to communicate with the device.
            _args (list): A list of arguments passed to the function.
            _kwargs (dict): A dictionary of keyword arguments passed to the function.
    """
    def __init__(self, session: object, *args, **kwargs):
        super().__init__(session)
        self._args = args
        self._kwargs = kwargs

    def _set_default_prompts(self):
        """
        Set the default prompts.
        """
        self._s.set_default_prompts([PROMPT_CONFIG])

    def _getin(self):
        """
        Get into the interface lan.
        """
        self._s.command_expect(f'router rip')

    def getin(self):
        """
        Set the default prompts and get into the interface wan.

            Returns: 
                self: The instance of the object. 
        """
        self._set_default_prompts()
        self._getin()
        return self

    def set_rip(self, network: list = None, version: str = "2", redistribute: str = None ):
        """
         Set RIP settings for the given network version and redistribute. This is useful for testing purposes and to ensure that you don't accidentally have a race condition in your database that is the user's responsibility.
         
            Args:
                network: The network to set the RIP for.
                version: The RIP version to set. If not specified the default is used.
                redistribute: The redistribute option to set. If not specified the default is used.
            
            Returns: 
                The command self for chaining.
        """
        
        # configure network if network is set
        if network:
            for iface in network:
                self._s.command_expect(f'network {iface}')
        # configure version if configure is set
        if version:
            self._s.command_expect(f'version {version}')
        # configure redistribute if redistribute is set
        if redistribute:
            self._s.command_expect(f'redistribute {redistribute}')
        return self


from scapy.all import Ether, IP, UDP, sendp

eth_frame = Ether( dst="ff:ff:ff:ff:ff:ff")

ip_packet = IP(src="{sip}", dst="{dip}")

udp_packet = UDP(dport={dport}) / "Broadcast message yaaaaa"

packet = eth_frame / ip_packet / udp_packet

sendp(packet, iface="{iface}", count={count})


import re
import os
import yaml
import time
import serial
import logging
import pexpect

from mdc import *
from pexpect import pxssh, TIMEOUT, EOF
from utils.debug_output import create_debug_output
from mdc.router.cli.common_func import telnet_login

lib_path = os.path.abspath(os.environ['PYTHONPATH'])
with open(f'{lib_path}/prompt.yaml') as f:
    PROMPTS = yaml.load(f, Loader=yaml.FullLoader)

class SshSession:
    """
    Class for managing SSH sessions.

        Args:
            host (str): Hostname of the server.
            port (int, optional): Port of the server. Defaults to 22.
            timeout (int, optional): Timeout for prompt waiting. Defaults to 2.

        Attributes:
            _ssh (pxssh): pxssh object for managing the session.
            _host (str): Hostname of the server.
            _port (int): Port of the server.
            _timeout (int): Timeout for prompt waiting.
    """

    def __init__(self, conf: dict, error_detector = None, auto_prompt_reset: bool = False):
        self._ssh = pxssh.pxssh(options={'StrictHostKeyChecking': 'no'})
        import sys
        self._ssh.logfile = sys.stdout.buffer  # debug
        self.conf = conf
        self._port = self.conf['port']
        self._host = self.conf['host']
        self._error_detector = error_detector
        self.auto_prompt_reset = auto_prompt_reset
        self._timeout = self.conf['timeout']
        self.reset_default_prompts()

    def is_connected(self):
        """
        Check if session is connected.
        """
        return self._ssh.isalive()

    def login(self, username: str, password: str):
        """
        Login to the server with given credentials.
        """
        try:
            self._ssh.login(self._host,
                            username,
                            password,
                            port=self._port,
                            login_timeout=120,
                            sync_original_prompt=False,
                            original_prompt=r"[#$>>]",
                            auto_prompt_reset=self.auto_prompt_reset)
        except pxssh.ExceptionPxssh as e:
            raise RuntimeError("can not login with SSH") from e

    def logout(self):
        """
        Logout from the server.
        """
        self._ssh.logout()

    def login_as_root(self):
            """
            Log in as root using the given session and sudo password.

                Args:
                    session (object): The session object to use for logging in.
                    sudo_paswd (str): The sudo password to use for logging in.
            """
            print('login_as_root')  # debug
            self._ssh.sendline('su')
            self._ssh.expect('Password')
            self._ssh.sendline(self.conf['credential']['password'])

    def logout_from_root(self):
        """
        Log out from root using the given session.

            Args: 
                session (object): The session object to use for logging out.
        """
        print('logout_from_root')  # debug
        self._ssh.sendline('exit')
        self._ssh.expect_exact('$')

    def sendcontrol(self, char: str):
        """
        Send control character to the server.
        """
        self._ssh.sendcontrol(char)

    def clear_buffer(self):
        """
        Clear buffer of the session. 
        """
        # ref https://github.com/pexpect/pexpect/issues/227
        self._ssh.expect_exact(self._ssh.buffer)

    def set_default_prompts(self, prompts: list):
        """
        Set the default prompts for the application.
        """
        self._prompts = prompts
    
    def reset_default_prompts(self):
        """
        Reset the default prompts for the application to all available prompts.
        """
        self._prompts = PROMPTS["ALL"]["MAIN"]

    def _check_session(self):
        """
        Check if session is opened. If not, raise ValueError.
        """
        if self._ssh is None:
            raise ValueError('session not opened')
        
    def close(self):
        """
        Close the connection
        """
        self._check_session()
        self._ssh.close()

    def command(self,
                cmd: str,
                parser=None,
                exact_prompts: list = None,
                timeout: int = None):
        """
        Send a command to the remote host and return the output.

            Args:
                cmd (str): The command to be executed.
                parser (function, optional): A function to parse the output of the command. Defaults to None. 
                exact_prompts (list, optional): List of expected prompts from the remote host after sending the command. Defaults to None. 
                timeout (int, optional): Timeout for waiting for a prompt from the remote host after sending the command. Defaults to None and uses self._timeout instead. 
                
            Returns: 
                str or function: Output of the command or result of parser if specified. Returns None if no prompt is matched within timeout period.
        """
        self.clear_buffer()
        to = timeout or self._timeout
        self._ssh.sendline(cmd)
        if exact_prompts is None:
            matched = self._ssh.prompt(timeout=to)
        else:
            try:
                self._ssh.expect_exact(exact_prompts, timeout=to)
                # self._ssh.expect_exact(exact_prompts, timeout=to)
            except (TIMEOUT, EOF):
                matched = False
            else:
                matched = True

        if not matched:
            print('can not match any prompt')
            # return None

        output = self._ssh.before.decode('utf-8')
        if parser is None:
            return output
        return parser(output)

    def command_expect(self,
                       command: str = None,
                       parser=None,
                       prompts: list = None,
                       timeout=None,
                       exact_str: bool = False):
        """
        Send a command to the session and expect a prompt. 
            Args: 
                command (str): The command to be sent to the session. 
                parser (function): A function to parse the output of the command. 
                prompts (list): A list of possible prompts that will be expected from the session. 
                timeout (int): Timeout for waiting for a prompt from the session. 
                exact_str (bool): Whether to expect an exact string or not. 

            Returns: 
                dict: A dictionary containing whether a prompt was matched, which pattern was matched and parsed data from output of command if parser is provided.
        """
        self.clear_buffer()
        if command is not None:
            _cmd = command + self.conf['terminator']
            self._ssh.send(_cmd)

        _prompts = prompts or self._prompts
        _prompts = [pexpect.EOF, pexpect.TIMEOUT] + _prompts

        _timeout = timeout or self.conf['timeout']

        if exact_str:
            idx = self._ssh.expect_exact(_prompts, timeout=_timeout)
        else:
            idx = self._ssh.expect(_prompts, timeout=_timeout)

        # strip first line because it is user-entered command
        # outputs[0] will be user-entered command
        # outputs[1] will be the return value of user-entered command
        outputs = self._ssh.before.decode('utf-8')
        outputs = outputs.split('\n', 1)
        # # debug section

        # print('\n-----------session.before-----------', f'cmd: "{command}"')
        # print(self._ssh.before)
        # print('\n------------------------------------')

        # # sometimes it won't return the correct data because the data has been split to session.after
        # # so maybe check what is in self._ssh.after

        # print('\n-----------session.after-----------', f'cmd: "{command}"')
        # print(self._ssh.after)
        # print('\n------------------------------------')
        output = ''
        if len(outputs) == 2:
            output = outputs[1]
        if self._error_detector is not None:
            self._error_detector(output)
            if isinstance(self._ssh.after,str):
                self._error_detector(self._ssh.after)

        return {
            'matched': idx > 1,
            'pattern': _prompts[idx],
            'data': output if parser is None else parser(output)
        }

class TelnetSession:
    """
    Constructs a CLI session object.

    Args:
        args (dict): Dictionary containing the following fields: 
        con_type (str): Connect type.
        host (str): Host's ip or domain name.
        port (str): Host's telnet port.
        prompts (list of strs): Prompts for check if connects.
        encode (str): Encode method.
        terminator (str): Line terminator.
        timout (int): Timeout for session. 

        error_detector (function, optional): Error detector function to be used in the session, defaults to None. 
        logfile (file, optional): Log file to be used in the session, defaults to None. 
    """

    def __init__(self, args: dict, error_detector=None, debug_output: bool = False):
        logging.info(f'args: {args}')
        if not set(args.keys()) >= {
                'con_type', 'host', 'port', 'prompts', 'encode', 'terminator',
                'timeout'
        }:
            raise ValueError('missing argument fields')
        self._args = args
        self._prompts = None
        self._session = None
        self._debug_output = debug_output
        self._error_detector = error_detector
        self._con_type = self._args["con_type"]
        if self._con_type not in ('ssh', 'telnet'):
            raise ValueError('bad con_type')
        self.reset_default_prompts()

    def set_default_prompts(self, prompts: list):
        """
        Set the default prompts for the application.
        """
        self._prompts = prompts

    def get_default_prompts(self):
        """
        Get the default prompts for the application.
        """
        return self._prompts

    def reset_default_prompts(self):
        """
        Reset the default prompts for the application to all available prompts.
        """
        self._prompts = PROMPTS["ALL"]["MAIN"]
        # self._prompts = self._args['prompts']

    def open(self) -> bool:
        """
        Open a connection to a remote host.

            Args:
                username (str): The username to use for authentication.
                password (str): The password to use for authentication.
                
            Returns: 
                bool: True if the connection was successful, False otherwise. 
        """
        self._session = pexpect.spawn(
            f'{self._args["con_type"]} \
            {self._args["host"]} {self._args["port"]}',
            timeout=self._args["timeout"],
            encoding=self._args["encode"],
        )

        self.reset_default_prompts()
        if self._debug_output:
            logfile = create_debug_output(self._debug_output, 'cli_session_edrg9010',
                                                self._con_type, self._args["host"], 
                                                self._args["port"])
            self._session.logfile = logfile

    def _router_telnet_login(self, username: str, password: str,
                             login_prompt: str):
        """
        Log into the router via telnet.

        Args:
            username (str): The username of the router.
            password (str): The password of the router.
            login_prompt (str, optional): The login prompt of the router. If not specified, all prompts will be used. 
        """
        # disable error detector tmeporarily because login would have ^]
        func_hold = self._error_detector
        self._error_detector = None
        if login_prompt is None:
            telnet_login(self, username, password, PROMPTS["ALL"]["MAIN"], PROMPTS["ALL"]["CONFIG"])
        else:
            telnet_login(self, username, password, PROMPTS["ALL"]["MAIN"], PROMPTS["ALL"]["CONFIG"], login_prompt)
        self._error_detector = func_hold

    # FIXME: arg "telnet_login_prompt" is a debt of fix TAS-25/27
    def login(self, credential: dict, telnet_login_prompt: str = None):
        """
        Log in to the router using the provided credentials.

        Args:
            credential (dict): Dictionary containing username and password.
            telnet_login_prompt (str, optional): Prompt used for telnet login. Defaults to None.
        """
        self._check_session()

        try:
            username = credential['username']
            password = credential['password']
        except KeyError as e:
            raise ValueError('Bad credential format') from e

        self._router_telnet_login(username, password, telnet_login_prompt)

    def logout(self):
        """
        Log out of the current session. 
        If the connection type is SSH, log out using the SSH session. 
        Otherwise, send an 'exit' command to the remote host. 

        """
        if self._con_type == 'ssh':
            self._session.close()
        else:
            self.command_expect('exit')

    def _check_session(self):
        """
        Check if session is opened. If not, raise ValueError.
        """
        if self._session is None:
            raise ValueError('session not opened')

    def clear_buffer(self):
        """
        Clear buffer of the session. 
        """
        # ref https://github.com/pexpect/pexpect/issues/227
        s = self._session
        s.expect_exact(s.buffer)

    def command_expect(self,
                       command: str = None,
                       parser=None,
                       prompts: list = None,
                       timeout=None,
                       exact_str: bool = False):
        """
        Send a command to the session and expect a prompt. 
            Args: 
                command (str): The command to be sent to the session. 
                parser (function): A function to parse the output of the command. 
                prompts (list): A list of possible prompts that will be expected from the session. 
                timeout (int): Timeout for waiting for a prompt from the session. 
                exact_str (bool): Whether to expect an exact string or not. 

            Returns: 
                dict: A dictionary containing whether a prompt was matched, which pattern was matched and parsed data from output of command if parser is provided.
        """
        self._check_session()

        if command is not None:
            _cmd = command + self._args['terminator']
            self._session.send(_cmd)

        _prompts = prompts or self._prompts
        _prompts = [pexpect.EOF, pexpect.TIMEOUT] + _prompts

        _timeout = timeout or self._args['timeout']

        if exact_str:
            idx = self._session.expect_exact(_prompts, timeout=_timeout)
        else:
            idx = self._session.expect(_prompts, timeout=_timeout)

        # strip first line because it is user-entered command
        # outputs[0] will be user-entered command
        # outputs[1] will be the return value of user-entered command
        outputs = self._session.before.split('\n', 1)

        # # debug section

        # print('\n-----------session.before-----------', f'cmd: "{command}"')
        # print(self._session.before)
        # print('\n------------------------------------')

        # # sometimes it won't return the correct data because the data has been split to session.after
        # # so maybe check what is in self._session.after

        # print('\n-----------session.after-----------', f'cmd: "{command}"')
        # print(self._session.after)
        # print('\n------------------------------------')
        output = ''
        if len(outputs) == 2:
            output = outputs[1]

        if self._error_detector is not None:
            self._error_detector(output)
            if isinstance(self._session.after,str):
                self._error_detector(self._session.after)

        return {
            'matched': idx > 1,
            'pattern': _prompts[idx],
            'data': output if parser is None else parser(output)
        }

    def command(self, command: str):
        """
        Commands to the host.

            Args:
                command: str, the text to command
        """
        self._check_session()

        self._session.send(command + self._args["terminator"])
        logging.info(command)

    def expect(self, prompts: list) -> dict:
        """
        Expect the specific response. The first two match are EOF & timeout.
        Return a dict collects the raw context and the matched prompt.

            Args:
                prompts: list of string,  the list for expect string.

            Returns:
                dict, format as
                {
                    "context": the raw response text from device,
                    "match_prompt": the prompt matched in the response context,
                    it could be EOF or TIMEOUT
                }
        """
        self._check_session()

        _prompts = [pexpect.EOF, pexpect.TIMEOUT] + prompts
        index = self._session.expect(_prompts)
        logging.debug(self._session.before)

        return {
            "context": self._session.before,
            "match_prompt": _prompts[index]
        }

    def close(self):
        """
        Close the connection
        """
        self._check_session()

        self._session.close()

    def sendcontrol(self, char: str):
        """
        Send control character to the remote host.

            Args: 
                char (str): The control character to send.
        """
        self._session.sendcontrol(char)

class SerialSession:
    """
    Constructs a CLI session object.

    Args:
        args (dict): Dictionary containing the following fields: 
        con_type (str): Connect type.
        host (str): Host's ip or domain name.
        port (str): Host's telnet port.
        prompts (list of strs): Prompts for check if connects.
        encode (str): Encode method.
        terminator (str): Line terminator.
        timout (int): Timeout for session.
        serial_port (str): Port number.
        baudrate (int): Serial port speed.

        error_detector (function, optional): Error detector function to be used in the session, defaults to None. 
        logfile (file, optional): Log file to be used in the session, defaults to None. 
    """

    def __init__(self, args: dict, error_detector=None, debug_output: bool = False):
        logging.info(f'args: {args}')
        if set(args.keys()) != {
                'con_type', 'host', 'port', 'prompts', 'encode', 'terminator',
                'timeout', 'serial_port', 'baudrate'
        }:
            raise ValueError('missing argument fields')
        if args["serial_port"] == None:
            raise ValueError('missing serial_port value')
        self._args = args
        self._prompts = None
        self._session = None
        self._debug_output = debug_output
        self._error_detector = error_detector
        self._con_type = self._args["con_type"]
        self.serial_port = self._args["serial_port"]
        self.baudrate = self._args["baudrate"]
        self.reset_default_prompts()

    def set_default_prompts(self, prompts: list):
        """
        Set the default prompts for the application.
        """
        self._prompts = prompts

    def get_default_prompts(self):
        """
        Get the default prompts for the application.
        """
        return self._prompts

    def reset_default_prompts(self):
        """
        Reset the default prompts for the application to all available prompts.
        """
        self._prompts = PROMPTS["ALL"]["MAIN"]
        # self._prompts = self._args['prompts']

    def open(self) -> bool:
        """
        Open a connection to a remote host.

            Returns: 
                bool: True if the connection was successful, False otherwise. 
        """
        self._session = serial.Serial(self.serial_port, baudrate=self.baudrate, timeout=1)

        self.reset_default_prompts()
        # if self._debug_output:
        #     logfile = create_debug_output(self._debug_output, 'cli_session_edrg9010',
        #                                         self._con_type, self._args["host"], 
        #                                         self._args["port"])
        #     self._session.logfile = logfile

    def login(self, credential: dict, telnet_login_prompt: str = None):
        """
        Log in to the router using the provided credentials.

        Args:
            credential (dict): Dictionary containing username and password.
            telnet_login_prompt (str, optional): Prompt used for telnet login. Defaults to None.
        """
        self._check_session()

        try:
            username = credential['username']
            password = credential['password']
        except KeyError as e:
            raise ValueError('Bad credential format') from e

        assert self._session.isOpen()
        self._session.reset_input_buffer()

        quick_timeout = 2

        # if login session existed, send enter to let prompt appear
        # print('try to get main prompt') #debug
        ret = self.command_expect('', prompts=PROMPTS["ALL"]["MAIN"], timeout=quick_timeout)
        # consume prompts to make buffer clean
        # ret = session.command_expect(None, prompts=promtps, timeout=quick_timeout)
        if ret['matched']:
            return

        # print('try to get config prompt') #debug
        # if login session is config prompt, try to back to main prompt
        ret = self.command_expect('', prompts=PROMPTS["ALL"]["CONFIG"], timeout=quick_timeout)
        if ret['matched']:
            # print('config mode') #debug
            is_main_prompt = False
            for i in range(120):
                # print('try to login') #debug
                time.sleep(1)
                is_main_prompt = self.command_expect('exit', prompts=PROMPTS["ALL"]["MAIN"], timeout=quick_timeout)['matched']
                if is_main_prompt:
                    break
                # print('exit to main prompt') #debug
            return

        self.command_expect('', prompts=["login: "])
        self.command_expect(username, prompts=["Password: "])
        ret = self.command_expect(password)
        if not ret['matched']:
            raise ValueError(f'cannot find prompt after password: {ret}')


    def logout(self):
        """
        Log out of the current session. 
        If the connection type is SSH, log out using the SSH session. 
        Otherwise, send an 'exit' command to the remote host. 

        """
        if self._con_type == 'ssh':
            self._session.logout()
        else:
            self.command_expect('exit')

    def _check_session(self):
        """
        Check if session is opened. If not, raise ValueError.
        """
        if self._session is None:
            raise ValueError('session not opened')

    def clear_buffer(self):
        """
        Clear buffer of the session.
        Clear both the input and output buffers of a serial port. 
        It discards any received but unread data from the input buffer and any unsent data from the output buffer. 
        Calling flush() ensures that all pending data is cleared before proceeding. 
        """
        s = self._session
        s.flush()

    def command_expect(self,
                       command: str = None,
                       parser=None,
                       prompts: list = None,
                       timeout=5,
                       exact_str: bool = False):
        """
        Send a command to the session and expect a prompt. 
            Args: 
                command (str): The command to be sent to the session. 
                parser (function): A function to parse the output of the command. 
                prompts (list): A list of possible prompts that will be expected from the session. 
                timeout (int): Timeout for waiting for a prompt from the session. 
                exact_str (bool): Whether to expect an exact string or not. 

            Returns: 
                dict: A dictionary containing whether a prompt was matched, which pattern was matched and parsed data from output of command if parser is provided.
        """
        prompts = prompts or self._prompts
        # print("command: ", command, prompts)
        print("command: ", command)
        if not isinstance(timeout, int):
            timeout = 5

        self._check_session()
        # print(f"*write: {command}")
        self._session.reset_input_buffer()
        self._session.write(f"{command}\n".encode())
        response = ""
        start_time = time.time()
        idx = -1

        while True:
            response += self._session.read(self._session.in_waiting or 1).decode()
            if exact_str:
                idx = next((i for i, prompt in enumerate(prompts) if response.endswith(prompt)), -1)
            else:
                idx = next((i for i, prompt in enumerate(prompts) if re.search(prompt, response)), -1)

            if idx >= 0:
                break

            if time.time() - start_time > timeout:
                break

        if f'{command}\r' == response.split('\n',1)[0]:
            output = response.split('\n',1)[1]
        else:
            output = response

        # print("*response: ",response,"end*")

        if self._error_detector is not None:
            self._error_detector(output)

        ret= {
            'matched': idx >= 0,
            'pattern': prompts[idx] if idx >= 0 else None,
            'data': output if parser is None else parser(output)
        }
        print(ret)
        return ret

    def command(self, command: str):
        self.command_expect(command)
    #     """
    #     Commands to the host.

    #         Args:
    #             command: str, the text to command
    #     """
    #     self._check_session()

    #     self._session.send(command + self._args["terminator"])
    #     logging.info(command)


    def close(self):
        """
        Close the connection
        """
        self._check_session()
        self._session.close()

    def sendcontrol(self, letter: str):
        """
        Send control character to the remote host.

            Args: 
                char (str): The control character to send.
        """
        ascii_value = ord(letter.upper())
        ctrl_ascii_value = ascii_value & 0b00011111
        ctrl_char = bytes([ctrl_ascii_value])
        self._session.write(ctrl_char)


import re
import sys
import time
import json

from host import *
from behave import *
from steps import syslog


def check_message(ret, msg):
    for r in ret:
        if msg in r['event']:
            return True
    return False


@given('"{device}" records event log "{message}"')
def step_impl(context, device, message):
    ret = context.dut[device].main().show_logging_event_log_system()
    assert check_message(ret, message), \
    f"DUT fail to record event logs: expect {message}, but {[r['event'] for r in ret]}"
    context.dut[device].main().clear(event_log='all')


@then('"{device}"s event log is cleared')
def step_impl(context, device):
    ret = context.dut[device].main().show_logging_event_log_system()
    assert ret[0]['number'] == '0', \
    f"DUT fail to clear system event logs: expect 0 number of event logs, but got {ret['number']} event logs"


@when('export "{device}" {category} eventlog file to {server} server on "{host}"')
def step_impl(context, device, category, server, host):
    context.dut[device].main().export_eventlog(server=server,
                                               server_ip=context.host_info[host]['testbed']['cur_host'],
                                               category=category)
 
@then(
    'the comparison between running {event_type} eventlog on "{device}" and exported system eventlog must be the same'
)
def step_impl(context, event_type, device):
    with open(f'//tftp/{event_type}.json') as f:
        data = json.load(f)
    print('data: ', data)
    ret = context.dut[device].main().show_logging_event_log_system()
    system_eventlog = dict(reversed(list(ret[0].items())))
    print('system_eventlog: ', system_eventlog)
    for i in range(max(len(system_eventlog), len(data[0]))):
        assert system_eventlog['event'] == data[0]['LOGSTR'], \
            f"from line {i} running system is different from ini:\n \
              running system line {i}: {system_eventlog['event']},\n\
              ini line {i}: {data[0]['LOGSTR']}"



from atb.service import *

def list_control_type_names():
    """
        list_control_type_names() -> list
        Return a list of control type names from the CONTROL_TYPE_MAP.
    """
    return list(CONTROL_TYPE_MAP)


def get_control_class(type_name):
    """
    Returns the control class associated with the given type name.

        Parameters:
            type_name (str): The name of the control type.

        Returns:
            class: The control class associated with the given type name. 
    """
    return CONTROL_TYPE_MAP[type_name]


CONTROL_TYPE_MAP = {
    'shellcmd':    ShellCmdService,
    'hping3':      Hping3Service,
    'tftp':        TftpService,
    'trex':        TrexService,
    'tshark':      TsharkService,
    'tcp_tool':    PktMgmtService,
    'syslog':      SyslogService,
    'trex_console':TrexConsoleService,
    'trex_server': TrexServerService,
    'scapy':       ScapyService,
    'dut_ui':      UICtrlService,
    'ca':          CaService,
    'bird':        BirdService,
    'nmcli':       NetworkManagerService,
    'radius':      RadiusService, 
    'snmp':        SnmpService,
    'mail_server': MailServerService
}


import time

from behave import *


@fixture
def getSyslogServer(context, host):
    try:
        context.hosts[host].syslog.start()
        yield
    finally:
        try:
            context.hosts[host].syslog.stop()
        except:
            print('syslog server is no activated')


@given('run syslog service on "{host}"')
def step_impl(context, host):
    use_fixture(getSyslogServer, context, host)


@given('set and enable syslog server with {host} IP on "{device}"')
def step_impl(context, device, host):
    print("syslog ip", context.host_info[host]["testbed"]["cur_host"])
    context.dut[device].go_config().set_syslog_server(
        action="create", ip=context.host_info[host]["testbed"]["cur_host"])
    context.dut[device].main().show_syslog_setting()


@given('"{device}" enable system-event with "{item}" for syslog action')
def step_impl(context, device, item):
    context.dut[device].go_config().set_warning_notification(
        event=item, active=True, action="Syslog only")


@then('"{host}" receive syslog "{receive_syslog}"')
def step_impl(context, host, receive_syslog):
    time.sleep(5)
    t = 0
    logs = context.hosts[host].syslog.read_message()
    while receive_syslog not in logs and t != 60:
        logs = context.hosts[host].syslog.read_message()
        time.sleep(1)
        t += 1
        # print(type(logs))
        print("logs=", logs)
    assert receive_syslog in logs, f'cannot find "{receive_syslog}" in syslog: {logs}'



from host import *
from behave import *
from steps import login, interface, common

@then('compare running config with exported configuration file')
def step_impl(context):
    with open(f'tftp/test.ini') as f:
        lines = f.read().replace(" ", "").split('\n')[:-1]
    for i in range(len(context.running_config)):
        assert context.running_config[i] == lines[
            i], f"from line {i} running config is different from ini"



from session import SshSession
from utils.decorator import root_access_required


class NetworkManagerService:
    """
    A class to manage NetworkManager server.

        Attributes:
            _ssh (SshSession): An instance of SshSession.
            _sudo_pwd (str): Password for sudo user.
    """

    def __init__(self, ssh_session: SshSession, sudo_paswd: str):

        self._ssh = ssh_session
        self._sudo_pwd = sudo_paswd

    def _check_conneted(self):
        """
        Check if SSH session is connected.

        Raises:
            ValueError: If SSH session is not connected.
        """
        if not self._ssh.is_connected():
            raise ValueError('SSH session not connected')

    def check_env(self):
        """
        Check the environment for the current SSH connection.

        Raises:
            ValueError: If the NetworkManage tool is not found.
        """
        self._check_conneted()

        ret = self._ssh.command('nmcli --version', timeout=5)
        assert 'not found' not in ret, \
                f'Command "nmcli" not found'

    @root_access_required
    def intf_setting(self, managed: bool = True):
        """
        Configure interface management status.

            Attributes:
                managed (bool): If managed is set to false, all interfaces listed in /etc/network/interfaces will be ignored.
        """
        self._check_conneted()

        if managed:
            cmd = "sed -i 's/managed=false/managed=true/g' /etc/NetworkManager/NetworkManager.conf"
        else:
            cmd = "sed -i 's/managed=true/managed=false/g' /etc/NetworkManager/NetworkManager.conf"

        self._ssh.command(cmd)

        ## restart
        cmd = 'systemctl restart NetworkManager'
        self._ssh.command(cmd, timeout=5)

    @root_access_required
    def create_connection(self,
                          con_type: str,
                          con_name: str = None,
                          vpn_type: str = None,
                          user: str = None,
                          password: str = None,
                          gateway: str = '192.168.128.254'):
        """
        Activate the network connection. It will take about 5 sec.

            Attributes:
                con_type (str): Connection type.
                                Type could be: ethernet, wifi, vpn, ppp, bridge, vlan, ...
                con_name (str): Connection name. Defaults to None.
                vpn_type (str): The type of VPN connection. Defaults to None.
                user (str): Username of the connection. Defaults to None.
                password (str): Password of the connection. Defaults to None.
                gateway (str): Connection gateway. Default to 192.168.128.254

            Raises: 
                Exception: If connection fails to create.
        """
        self._check_conneted()

        cmd = 'nmcli connection add '
        cmd += f'type {con_type} '
        cmd += f'con-name "{con_name}" '

        if vpn_type:
            cmd += f'vpn-type "{vpn_type}" '

        cmd += f'vpn.data "gateway={gateway}, user={user}" '
        cmd += f'vpn.secrets "password={password}" '

        ret = self._ssh.command(cmd, timeout=10)

        if 'successfully added' not in ret:
            raise ValueError(f'Connection create fail, {ret}')

    @root_access_required
    def activate_connection(self, con_name: str, ifname: str = 'eth1'):
        """
        Activate the network connection.

            Attributes:
                con_name (str): Connection to be Activated.

            Raises: 
                Exception: If connection fails to activate.
        """
        self._check_conneted()

        cmd = f'nmcli connection up {con_name} '
        cmd += f'ifname {ifname}'

        ret = self._ssh.command(cmd, timeout=15)

        if 'successfully activated' not in ret and 'already active' not in ret:
            raise ValueError(f'Connection activate fail, {ret}')

    @root_access_required
    def deactivated_connection(self, con_name: str):
        """
        Deactivate the network connection.

            Attributes:
                con_name (str): Connection to be deactivated.

            Raises: 
                Exception: If connection fails to deactivate.
        """
        self._check_conneted()

        cmd = f'nmcli connection down {con_name}'

        ret = self._ssh.command(cmd, timeout=10)

        if 'successfully deactivated' not in ret:
            raise ValueError(f'Connection deactivate fail, {ret}')

    @root_access_required
    def delete_connection(self, con_name: str):
        """
        Delete the network connection.

            Attributes:
                con_name (str): Connection to be deleted.

            Raises: 
                Exception: If connection fails to delete.
        """
        self._check_conneted()

        cmd = f'nmcli connection delete {con_name}'

        ret = self._ssh.command(cmd, timeout=10)

        if 'successfully deleted' not in ret:
            raise ValueError(f'Connection delete fail, {ret}')

    def show_connection(self):
        """
        Show all the network connection.
        """
        self._check_conneted()

        self._ssh.command('nmcli connection show', timeout=5)



import re
import time
import requests
from requests.auth import HTTPBasicAuth

from behave import *
from host import HostFactory, SessionFactory
from steps.login import loginToDUT
from topology.topology_info import *
from utils.threading import ThreadWithResult
from mdc.router.cli.rp_base.lg_router.tn5000.base import Base as tn5000

def login_and_measure_time(context, device):
    # try to login DUT after restart
    elapsed_time = 0
    start_time = time.time() 
    is_login = False
    while not is_login and elapsed_time <= 120:
        try:
            end_time = time.time()
            elapsed_time = end_time - start_time
            print('is_login: ', is_login)
            is_login = loginToDUT(context, device)
        except Exception as e:
            print(f'Can not login to DUT yet: {e}')
    
    return elapsed_time

@given(u'prepare "{file_name}" topology')
def step_impl(context, file_name):
    ''' get topology file absloute path from lib 
        e.g. 
            cur_path = /usr/src/app/sqa/{function}/
            root_path = /usr/src/app/
            abs_path = /usr/src/app/lib/topology/
    '''
    prepare_json(context, file_name)


@when('clear "{device}" all logging event log')
@given('clear "{device}" all logging event log')
def step_impl(context, device):
    if issubclass(context.dut[device]._model, tn5000):
        context.dut[device].main().clear_event_log()
    else:
        context.dut[device].main().clear(event_log="all")


def reset(context, device):
    # Step-1: authorize CLI of "{device}"
    # Step-2: reload factory-default "{device}"
    # Step-3: clear "{device}" all logging event log
    context.execute_steps(f'''given authorize CLI of "{device}"''')
    context.execute_steps(f'''when reload factory-default "{device}"''')
    context.execute_steps(f'''given clear "{device}" all logging event log''')


@given('reset all devices')
def step_impl(context):
    reset_threads = []
    for row in context.table:
        reset_threads.append(
            ThreadWithResult(target=reset,
                             kwargs={
                                 'context': context,
                                 'device': row['devices']
                             }))
    for t in reset_threads:
        t.start()
    for t in reset_threads:
        t.join()


@given(u'prepare a "{host}" to connect to "{device}"')
def step_impl(context, host, device):
    info = context.host_info[host]
    context.hosts[host] = HostFactory.create_atb(info)
    context.hosts[host].login()
    context.hosts[host].check_env()
    context.hosts[host].shellcmd.remove_residual_artifact()
    context.hosts[host].shellcmd.set_network(dev=info["nic"], 
                                             ip=info["testbed"]["lan_host"], 
                                             mask=info["testbed"]["mask"])
        
@given(u'check "{host}" has installed {tool}')
def step_impl(context, host, tool):
    if tool.lower() == "tshark":
        context.hosts[host].tshark.check_env()
    elif tool.lower() == "hping3":
        context.hosts[host].hping3.check_env()


@when('reload factory-default "{device}"')
@given('reload factory-default "{device}"')
def step_impl(context, device):
    def _backup():
        ret = context.dut[device].main().import_config(
            tftp_ip=context.host_info["HOST_EXECUTOR"]["testbed"]["cur_host"],
            file_name="default.ini")
        context.dut[device].close()
        assert ret['matched'] and '^Parse error' not in ret[
            'data'], f'Configuration Import Fail: {ret["data"]}'
        login_and_measure_time(context=context, device=device)
    def _reset():
        context.dut[device].go_config().set_hostname('init_check')
        sys_name = context.dut[device].main().show_system()[0]['system_name']
        assert sys_name == 'init_check', \
            f'Set system_name fail: {sys_name}'
        
        context.dut[device].main().reload_factory_default()
        time.sleep(context.dut_info[device]['reload_delay'])
        context.dut[device].close()
        

        login_time = login_and_measure_time(context=context, device=device)
        print(f"Reload Factory Time: {context.dut_info[device]['reload_delay']} + {login_time:.2f} seconds")

        sys_name = context.dut[device].main().show_system()[0]['system_name']
        assert "init_check" not in sys_name, \
            f'Bad hostanme after reload: {sys_name}'
    if context.dut_info[device]['init_method'] == "restore":
        _backup()
    elif context.dut_info[device]['init_method'] == "reset":
        _reset()
    else:
        raise ValueError(f"input wrong initial method, \
                         it should be reset or restore: {context.dut_info[device]['init_method']}")


def do_ping(context, machine1, machine2, result):
    '''
    parameter
        machine can be DUT or host(VM)
        there will be 4 combinations
        1. dut -> host 
        2. dut -> dut 
        3. host -> host 
        4. host -> dut 

        Note:
        if machine2 is dut, you need to specify interface for machine2
        e.g.
            "DUT" will ping "DUT WAN" successfully
            "HOST_A" will ping "DUT WAN" successfully
    '''
    if "dut" in machine1.lower() and "host" in machine2.lower():
        context.hosts[machine2].shellcmd.restart_network(context.host_info[machine2]["nic"])
        for i in range(5):
            # retry 5 times
            r = context.dut[machine1].main().ping(
                context.host_info[machine2]['testbed']['cur_host'])
            if int(r['received']) > 0:
                break
        if result == "successfully":
            assert int(
                r['received']
            ) == 4, f'expect {machine1} ping {machine2} {result}, Packets: Received = {int(r["received"])}'
        else:
            assert int(
                r['received']
            ) == 0, f'expect {machine1} ping {machine2} {result}, Packets: Received = {int(r["received"])}'
        # assert int(r['received']) > 0, f'{machine1} fail to ping {machine2}, Packets: Received = {int(r["received"])}'

    elif "dut" in machine1.lower() and "dut" in machine2.lower():
        machine2_list = machine2.split(' ')
        if len(machine2_list) == 2:
            machine2_name = machine2.split(' ')[0]
            machine2_intf = machine2.split(' ')[1]
        else:
            raise ValueError(
                f"format error: {machine2}, it should contain interface name")
        for i in range(5):
            # retry 5 times
            r = context.dut[machine1].main().ping(
                context.dut_info[machine2_name]["testbed"]
                [f"{machine2_intf.lower()}_ip"])
            if int(r['received']) > 0:
                break
        if result == "successfully":
            assert int(
                r['received']
            ) == 4, f'expect {machine1} ping {machine2} {result}, Packets: Received = {int(r["received"])}'
        else:
            assert int(
                r['received']
            ) == 0, f'expect {machine1} ping {machine2} {result}, Packets: Received = {int(r["received"])}'

    elif "host" in machine1.lower() and "host" in machine2.lower():
        context.hosts[machine1].shellcmd.restart_network(context.host_info[machine1]["nic"])
        context.hosts[machine2].shellcmd.restart_network(context.host_info[machine2]["nic"])
        for i in range(5):
            # retry 5 times
            r = context.hosts[machine1].shellcmd._ssh.command(
                cmd=
                f'ping {context.host_info[machine2]["testbed"]["cur_host"]} -c 4',
                timeout=10)
            # print(r)
            try:
                count = int(re.findall(r'(\d+)\s+received', r)[0])
            except:
                count = None
            if count == 4:
                break
        if result == "successfully":
            assert count == 4, f'expect {machine1} ping {machine2} {result}, Packets: Received = {count}'
        else:
            assert count == 0 or count == None, f'expect {machine1} ping {machine2} {result}, Packets: Received = {count}'

    elif "host" in machine1.lower() and "dut" in machine2.lower():
        context.hosts[machine1].shellcmd.restart_network(context.host_info[machine1]["nic"])
        machine2_list = machine2.split(' ')
        if len(machine2_list) == 2:
            machine2_name = machine2.split(' ')[0]
            machine2_intf = machine2.split(' ')[1]
        else:
            raise ValueError(
                f"format error: {machine2}, it should contain interface name")
        for i in range(5):
            # retry 5 times
            print(f"this is the {i+1} time!!!")
            r = context.hosts[machine1].shellcmd._ssh.command(
                cmd=
                f'ping {context.dut_info[machine2_name]["testbed"][f"{machine2_intf.lower()}_ip"]} -c 4',
                exact_prompts='packet loss',
                timeout=10)
            print(r)
            try:
                count = int(re.findall(r'(\d+)\s+received', r)[0])
            except:
                count = None
            if count == 4:
                break
        if result == "successfully":
            assert count == 4, f'expect {machine1} ping {machine2} {result}, Packets: Received = {count}'
        else:
            assert count == 0 or count == None, f'expect {machine1} ping {machine2} {result}, Packets: Received = {count}'
    else:
        if "host" in machine1.lower():
            context.hosts[machine1].shellcmd.restart_network(context.host_info[machine1]["nic"])
            for i in range(2):
                # retry 2 times
                print(f"this is the {i+1} time!!!")
                r = context.hosts[machine1].shellcmd._ssh.command(
                    cmd=
                    f'ping {machine2} -c 4',
                    exact_prompts='packet loss',
                    timeout=10)
                print(r)
                try:
                    count = int(re.findall(r'(\d+)\s+received', r)[0])
                except:
                    count = None
                if count == 4:
                    break
            if result == "successfully":
                assert count == 4, f'expect {machine1} ping {machine2} {result}, Packets: Received = {count}'
            else:
                assert count == 0 or count == None, f'expect {machine1} ping {machine2} {result}, Packets: Received = {count}'
        elif "dut" in machine1.lower():
            for i in range(2):
                # retry 2 times
                r = context.dut[machine1].main().ping(machine2)
                if int(r['received']) > 0:
                    break
            if result == "successfully":
                assert int(
                    r['received']
                ) == 4, f'expect {machine1} ping {machine2} {result}, Packets: Received = {int(r["received"])}'
            else:
                assert int(
                    r['received']
                ) == 0, f'expect {machine1} ping {machine2} {result}, Packets: Received = {int(r["received"])}'
        else:
            raise Exception("Suppose never goes here. Something wrong!")


def chk_arp(context, host_name=None, dut_name=None):
    if (host_name is None) and (dut_name is None):
        raise Exception("Empty input parameter for ARP check!")
    if host_name:
        context.hosts[host_name].shellcmd._ssh.command(
            cmd='export PATH="/usr/local/sbin:/usr/local/bin:/sbin:'
            '/bin:/usr/sbin:/usr/bin:/root/bin:/usr/local/games:'
            '/usr/games"',
            timeout=10)
        print("--------------check vm {host_name} arp table--------------")
        print(context.hosts[host_name].shellcmd._ssh.command(cmd="arp -n",
                                                        timeout=10))
        print("--------------check vm {host_name} arp table end --------------")
    if dut_name:
        print("--------------check dut {dut_name} arp table--------------")
        print(context.dut[dut_name].main().show_arp())
        print("--------------check dut {dut_name} arp table end --------------")


def clear_arp(context, clear_tgt_ip, host_name=None, dut_name=None):
    print('host_name' + str(host_name))
    if (host_name is None) and (dut_name is None):
        raise Exception("Empty input parameter for ARP check!")
    if host_name:
        print("--------------clear vm {machine1} arp table--------------")
        cmd = f'arp -D {clear_tgt_ip}'
        print(context.hosts[host_name].shellcmd._ssh.command(cmd, timeout=10))
    if dut_name:
        pass


def do_extra_after_ping(context, machine1, machine2, machine2_interface=None):
    '''
    parameter
        machine can be DUT or host(VM)
        there will be 4 combinations (MECE): 
        1. dut -> host 
        2. dut -> dut 
        3. host -> host 
        4. host -> dut 

        Note:
        if machine2 is dut, you need to specify interface for machine2
        e.g.
            "DUT" will ping "DUT WAN" successfully
            "HOST_A" will ping "DUT WAN" successfully
    '''
    if "dut" in machine1.lower() and "host" in machine2.lower():
        pass
    elif "dut" in machine1.lower() and "dut" in machine2.lower():
        pass
    elif "host" in machine1.lower() and "host" in machine2.lower():
        chk_arp(context, host_name=machine1)
    elif "host" in machine1.lower() and "dut" in machine2.lower():
        if machine2_interface:
            if "wan" in machine2_interface.lower():
                clear_arp(context,
                          context.dut_info[machine2]["testbed"]["wan_ip"],
                          host_name=machine1)
            elif "lan" in machine2_interface.lower():
                clear_arp(context,
                          context.dut_info[machine2]["testbed"]["lan_ip"],
                          host_name=machine1)
            elif "brg" in machine2_interface.lower():
                clear_arp(context,
                          context.dut_info[machine2]["testbed"]["brg_ip"],
                          host_name=machine1)
            else:
                raise Exception(
                    f"Non-Supported interface of machine2: {machine2_interface}"
                )
        chk_arp(context, host_name=machine1)
        chk_arp(context, dut_name=machine2)
    else:
        raise Exception(
            f"Suppose never goes here. Something wrong! machine1:{machine1}  machine2:{machine2}  machine2 interface:{machine2_interface}"
        )


@given('"{machine1}" will ping "{machine2}" {result}')
@then('"{machine1}" will ping "{machine2}" {result}')
def step_impl(context, machine1, machine2, result):
    '''
    parameter
        machine can be DUT or host(VM)
        there will be 6 combinations
        1. dut -> host 
        2. dut -> dut 
        3. host -> host 
        4. host -> dut 
        5. host -> any
        6. dut -> any

        Note:
        if machine2 is dut, you need to specify interface for machine2
        e.g.
            "DUT" will ping "DUT WAN" successfully
            "HOST_A" will ping "DUT WAN" successfully
        if you want to ping a specific address or URL, simply replace "machine2" with the desired address or URL.
        e.g.
            "DUT" will ping "google.com" successfully
            "HOST_A" will ping "192.168.127.253" successfully
    table
    '''
    try:
        for row in context.table:
            print('-----------------------------------------------')
            print(f'machine1: {row["device1"]}    machine2: {row["device2"]}')
            print('-----------------------------------------------')
            try:
                if context.dut[row['device1']]:
                    if "host" not in row['device1']:
                        context.dut[row['device1']].close()
                        loginToDUT(context, row['device1'])
                    print(f'{row["device1"]} already opened')
            except:
                if "host" in row['device1']:
                    context.hosts[row['device1']] = HostFactory.create_atb(
                        context.host_info[row['device1']])
                    context.hosts[row['device1']].login()
                    print('context.hosts' + str(context.hosts))
                else:
                    loginToDUT(context, row['device1'])

            do_ping(context, row['device1'], row['device2'], row['result'])
            # In case that machine2 is {DUT interface}, only DUT name is required.
            # extra steps after ping
            device1 = row['device1']
            device2_list = row['device2'].split(' ')
            device2 = device2_list[0]
            if len(device2_list) == 2:
                device2_interface = device2_list[1]
            else:
                device2_interface = None
            do_extra_after_ping(context, device1, device2, device2_interface)
    except:
        do_ping(context, machine1, machine2, result)

@then('{action} user account on "{device}" will {result}')
def step_impl(context, action, device, result):
    for row in context.table:
        if result == "fail":
            try:
                if 'privilege' in context.table.headings:
                    context.dut[device].go_config().set_login_account(
                        action=action,
                        username=row['username'],
                        password=row['password'],
                        privilege=row['privilege'])
                else:
                    context.dut[device].go_config().set_login_account(
                        action=action,
                        username=row['username'],
                        password=row['password'])
            except:
                print("OK! Suppose to catch exception!")
            else:
                assert False, f"{device} should not be able to {action} user account"
        elif result == "succeed":
            print(context.table.headings)
            if 'privilege' in context.table.headings:
                context.dut[device].go_config().set_login_account(
                    action=action,
                    username=row['username'],
                    password=row['password'],
                    privilege=row['privilege'])
            else:
                context.dut[device].go_config().set_login_account(
                    action=action,
                    username=row['username'],
                    password=row['password'])


@when('{action} user account on "{device}"')
def step_impl(context, action, device):
    username = password = privilege = None
    try:
        if context.table[0]['username']:
            username = context.table[0]['username']
        if context.table[0]['password']:
            password = context.table[0]['password']
        if context.table[0]['privilege']:
            privilege = context.table[0]['privilege']
    except:
        pass

    r = context.dut[device].go_config().set_login_account(username=username,
                                                          password=password,
                                                          privilege=privilege,
                                                          action=action)
    print(r)


@given('export "{device}" configuration file to {server}')
@when('export "{device}" configuration file to {server}')
def step_impl(context, device, server):
    ret = context.dut[device].main().export_config(
        tftp_ip=context.host_info["HOST_EXECUTOR"]["testbed"]["cur_host"])
    assert ret['matched'] and '^Parse error' not in ret[
        'data'], f'Configuration Upload Fail: {ret["data"]}'


@given('import {cfg_file} to "{device}" from {server}')
@when('import {cfg_file} to "{device}" from {server}')
def step_impl(context, device, server, cfg_file):
    ret = context.dut[device].main().import_config(
        tftp_ip=context.host_info["HOST_EXECUTOR"]["testbed"]["cur_host"],
        file_name=cfg_file)
    context.dut[device].close()
    assert ret['matched'] and '^Parse error' not in ret[
        'data'], f'Configuration Import Fail: {ret["data"]}'
    login_and_measure_time(context=context, device=device)


@given(
    'upgrade {firmware} to "{device}" from {server} and do warm start automatically on "{device}"'
)
@when(
    'upgrade {firmware} to "{device}" from {server} and do warm start automatically on "{device}"'
)
def step_impl(context, firmware, device, server):
    context.dut[device].main().save()
    ret = context.dut[device].main().upgrade_firmware(
        tftp_ip=context.host_info["HOST_EXECUTOR"]["testbed"]["cur_host"],
        file_name=firmware)
    context.dut[device].close()
    assert ret['matched'] and '^Parse error' not in ret['data']

    login_time = login_and_measure_time(context=context, device=device)
    print(f"Warm Start Time: {context.dut_info[device]['reload_delay']} + {login_time:.2f} seconds")
    time.sleep(10)


@given('get running config from "{device}"')
@when('get running config from "{device}"')
def step_impl(context, device):
    ret = context.dut[device].main().get_running_config(context.dut_info[device]["session"]["con_type"])
    context.running_config = ret.replace(" ", "").replace("\r", "").split('\n')
    context.running_config = list(filter(lambda x: x != '', context.running_config ))
    print("============context.running_config============")
    search_string = "showrunning-config"
    try:
        for index, item in enumerate(context.running_config):
            if search_string in item:
                context.running_config = context.running_config[index+1:]
                break
    except ValueError:
        print("No showrunning-config in running config")
    # if re.search(r'showrunning-config', context.running_config[0]) != None:
    #     context.running_config = context.running_config[1:]
    if re.search(r'[^\n]+[^\)]#', context.running_config[-1]) != None:
        context.running_config = context.running_config[:-1]
    if re.search(r'[^\n]+[^\)]>>', context.running_config[-1]) != None:
        context.running_config = context.running_config[:-1]
    

@then(
    'the comparison between running config and exported config must be the same'
)
def step_impl(context):
    with open('//tftp/test.ini') as f:
        lines = f.read().replace(" ", "").split('\n')[:-1]
    print('config: ', context.running_config)
    print('lines: ', lines)
    for i in range(max(len(context.running_config), len(lines))):
        assert context.running_config[i] == lines[i], \
            f"from line {i} running config is different from ini:\n \
              running config line {i}: {context.running_config[i]},\n\
              ini line {i}: {lines[i]}"


@then(
    '"{host}" tries to establish {ssh_num} ssh and {telnet_num} telnet connection to "{dut_name}"s {dut_itf} but only default {correct_num_of_session} session can be established'
)
@then(
    '"{host}" tries to establish {ssh_num} ssh and {telnet_num} telnet connection to "{dut_name}"s {dut_itf} but only modified {correct_num_of_session} session can be established'
)
def step_impl(context, host, ssh_num, telnet_num, dut_name, dut_itf,
              correct_num_of_session):
    context.dut[dut_name].main().clear_ssh_telnet_limit()

    # quit executor control session
    context.execute_steps(f'''given logout "{dut_name}"''')

    dut_ip = context.dut_info[dut_name]["testbed"][f"{dut_itf.lower()}_ip"]
    dut_account = context.dut_info[dut_name]["credential"]["username"]
    dut_password = context.dut_info[dut_name]["credential"]["password"]

    context.hosts[host].dut_ui.cli_init()
    for i in range(int(ssh_num)):
        context.hosts[host].dut_ui.cli_cmd(
            f'''/bin/bash -c 'sshpass -p {dut_password} ssh -o "StrictHostKeyChecking no" {dut_account}@{dut_ip} >> qqq.log' ''',
            detach=True)
        time.sleep(1)

    for i in range(int(telnet_num)):
        context.hosts[host].dut_ui.cli_cmd(
            f'''/bin/bash -c '(sleep 2; echo "{dut_account}"; sleep 2; echo "{dut_password}"; sleep 2) | telnet {dut_ip} >> qqq.log' ''',
            detach=True)
        time.sleep(3)
    for i in range(2):
        print(f'this is {i} time to cat session log')
        ret = context.hosts[host].dut_ui.cli_cmd("/bin/bash -c 'cat qqq.log'")
        login_symbol = 'MOXA'
        if login_symbol in ret:
            login_count = ret.count(login_symbol)
            break
    assert login_count == int(
        correct_num_of_session), f"Only {login_count} session(s)!"
    context.hosts[host].dut_ui.cli_destroy()

    # regain executor control
    context.execute_steps(f'''given login "{dut_name}" as Admin''')


@when('modify threshold of telnet and ssh connections on "{device}"')
def step_impl(context, device):
    threshold = int(context.table[0]['threshold'])
    context.dut[device].go_config().set_global_telnet(max_login_users=threshold)


@When('warm start on "{device}"')
def step_impl(context, device):
    context.dut[device].main().save()
    context.dut[device].main().reload()
    context.dut[device].close()
    time.sleep(context.dut_info[device]['reload_delay'])
    
    login_time = login_and_measure_time(context=context, device=device)
    print(f"Warm Start Time: {context.dut_info[device]['reload_delay']} + {login_time:.2f} seconds")


@When('cold start on "{device}"')
def step_impl(context, device):
    context.dut[device].main().save()
    if context.dut_info[device]["testbed"]["power_jig_device"] == "ip_power":
        requests.get(
            f'http://{context.dut_info[device]["testbed"]["power_jig_ip"]}/goform/setpower?p6{context.dut_info[device]["testbed"]["power_jig"]}=0',
            auth=HTTPBasicAuth('admin', '12345678'))
        time.sleep(1)
        requests.get(
            f'http://{context.dut_info[device]["testbed"]["power_jig_ip"]}/goform/setpower?p6{context.dut_info[device]["testbed"]["power_jig"]}=1',
            auth=HTTPBasicAuth('admin', '12345678'))
        context.dut[device].close()
        time.sleep(context.dut_info[device]['reload_delay'])
    
        login_time = login_and_measure_time(context=context, device=device)
        print(f"Cold Start Time: {context.dut_info[device]['reload_delay']} + {login_time:.2f} seconds")

    elif context.dut_info[device]["testbed"]["power_jig_device"] == "turtle":
        requests.get(
            f'http://{context.dut_info[device]["testbed"]["power_jig_ip"]}/outlet.cgi?outlet={context.dut_info[device]["testbed"]["power_jig"]}&command=2&time=1679423114457',
            auth=HTTPBasicAuth('admin', ''))
        time.sleep(5)
        requests.get(
            f'http://{context.dut_info[device]["testbed"]["power_jig_ip"]}/outlet.cgi?outlet={context.dut_info[device]["testbed"]["power_jig"]}&command=2&time=1679423114457',
            auth=HTTPBasicAuth('admin', ''))
        context.dut[device].close()
        time.sleep(context.dut_info[device]['reload_delay'])
    
        login_time = login_and_measure_time(context=context, device=device)
        print(f"Cold Start Time: {context.dut_info[device]['reload_delay']} + {login_time:.2f} seconds")
    else:
        raise ValueError(f"Wrong power_jig_device input: {context.dut_info[device]['testbed']['power_jig_device']}")
    


@given(u'set system name to "{name}" on "{device}"')
@when(u'set system name to "{name}" on "{device}"')
def step_impl(context, name, device):
    context.dut[device].go_config().set_hostname(name)


@Given('set default network on "LAN" interface and '
       'rename it as "{name}" on "{device}"')
@when('set default network on "LAN" interface and '
      'rename it as "{name}" on "{device}"')
def step_impl(context, name, device):
    # "DUT-1" set default network on "LAN" interface and rename it as "Lan-127"
    context.dut[device].go_config_if_lan().\
        ip_address_static(context.dut_info[device]['testbed']['lan_ip'], '255.255.255.0')
    context.dut[device].go_config_if_lan().name(name)

    r = context.dut[device].main().show_interface('lan')
    count = 0
    while context.dut_info[device]['testbed']['lan_ip'] not in r[0][
            'ip_address'] and count < 60:
        time.sleep(1)
        r = context.dut[device].main().show_interface('lan')
        count += 1


@Given('{action} trusted access on "{device}"')
@when('{action} trusted access on "{device}"')
def step_impl(context, device, action):
    time.sleep(1)
    context.dut[device].go_config().set_trusted_access(global_toggle=action)


@Given('save configuration into flash on "{device}"')
@when('save configuration into flash on "{device}"')
def step_impl(context, device):
    time.sleep(1)
    context.dut[device].main().save()


@Given('set and enable syslog server with "{host}" ip on "{device}"')
def step_impl(context, host, device):
    context.dut[device].go_config().set_syslog_server(
        'create', context.host_info[host]['testbed']['cur_host'])
    context.dut[device].main().show_logging_event_log_system()


@Given('enable system-event with "{item}" for syslog action on "{device}"')
def step_impl(context, item, device):
    context.dut[device].go_config().set_warning_notification(
        item, action='Syslog only', active=True)


@Given('enable system-event with "{item}" for "{event}" action on "{device}"')
def step_impl(context, item, event, device):
    context.dut[device].go_config().set_warning_notification(
        item, action=event, active=True)


@then('"{device}" record event log "{receive_log}"')
def step_impl(context, device, receive_log):
    logs = context.dut[device].main().show_logging_event_log_system()
    for log in logs:
        if receive_log in log['event']:
            result = 'pass'
            break
        else:
            result = 'fail'
    assert result == 'pass', f'cannot found "{receive_log}" in event log'

@given('connect to "{jig}"')
def step_impl(context, jig):
    print(context.jig_info[jig]["session"])
    context.jig = SessionFactory.create_telnet(context.jig_info[jig]["session"])


from mdc.router.cli.rp_base.lg_router.tn5000 import config
from mdc.router.cli.rp_base.lg_router.tn5000.cli import PROMPT_CONFIG

class ConfigObjectIndustrialAppServ(config.Config):
    """
    Class for configuring Industrial Application Service Object.

        Args:
            session (object): The session object.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Attributes:
            _s (Session): A Session object used to communicate with the device.
            _args (list): A list of arguments passed to the function.
            _kwargs (dict): A dictionary of keyword arguments passed to the function.
    """

    def __init__(self, session: object, *args, **kwargs):
        super().__init__(session)
        self._args = args
        self._kwargs = kwargs

    def _set_default_prompts(self):
        """
        Set the default prompts.
        """
        self._s.set_default_prompts([PROMPT_CONFIG])

    def _getin(self):
        """
        Get into the config-obj-indust-app prompt.
        """
        self._s.command_expect(f'object industrial-application-service')

    def getin(self):
        """
        Set the default prompts and get into the config-obj-indust-app prompt.

            Returns: 
                self: The instance of the object. 
        """
        self._set_default_prompts()
        self._getin()
        return self

    def create_object(self, name: str = None, service: list = None):
        """
        Create an object with the given name and service(s).

            Args:
                name (str, optional): The name of the object to create.
                service (list, optional): The service(s) of the object to create.
                
            Returns:
                self: The instance of the object. 
        """
        services = [
            "Modbus", "DNP3", "IEC-60870-5-104", "IEC-61850-MMS", "OPC-DA",
            "OPC-UA", "CIP-EtherNet/IP", "Siemens-Step7", "Moxa-RealCOM",
            "Moxa-MXview-Request"
        ]
        if name:
            self._s.command_expect(f'name {name}')
        if service:
            for i in service:
                if i in services:
                    self._s.command_expect(f'select {i}')
                else:
                    raise ValueError(f'input invalid service: {i}')

        # quit to save
        self._s.command_expect(f'exit')
        return self



import time

from behave import *

@then('the following rule should be on "{device}"\'s routing table')
def step_impl(context, device):
    flag = False
    for _ in range(5):
        ret = context.dut[device].main().show_ip_route()
        if any(
        item['type'] == context.table[0]['Type'] and
        item['destination'] == context.table[0]['Destination'] and
        item['next_hop'] == context.table[0]['Next Hop'] and
        item['interface'] == context.table[0]['interface']
        for item in ret): 
            flag = True
            break

    assert flag, f"No match found -> expect: {context.table[0]}, actual: {ret}"


from mdc.router.cli.rp_base.ng_router import cli, main, config, config_brg, config_nat, config_dhcp, config_ospf, \
    config_if_lan, config_if_wan, config_if_vlan, config_if_ether, \
    config_ipsec, config_ipsec_phase1, config_ipsec_phase2, \
    config_l2_filter, config_l3_policy, \
    config_obj_addr, config_obj_serv, config_obj_net_serv, config_obj_indust_app_serv, \
    config_vrrp, config_session_ctrl, config_rstp, config_rip

class Base:

    class Cli(cli.Cli):
        pass
    class Main(main.Main):
        pass
    class Config(config.Config):
        pass
    class ConfigIfLan(config_if_lan.ConfigIfLan):
        pass
    class ConfigIfWan(config_if_wan.ConfigIfWan):
        pass
    class ConfigIfVlan(config_if_vlan.ConfigIfVlan):
        pass
    class ConfigIfEthernet(config_if_ether.ConfigIfEthernet):
        pass
    class ConfigBrg(config_brg.ConfigBrg):
        pass
    class ConfigL2filter(config_l2_filter.ConfigL2filter):
        pass
    class ConfigIpsec(config_ipsec.ConfigIpsec):
        pass
    class ConfigIpsecPhase1(config_ipsec_phase1.ConfigIpsecPhase1):
        pass
    class ConfigIpsecPhase2(config_ipsec_phase2.ConfigIpsecPhase2):
        pass
    class ConfigL37Policy(config_l3_policy.ConfigL37Policy):
        pass
    class ConfigSessionControl(config_session_ctrl.ConfigSessionControl):
        pass
    class ConfigObjectAddr(config_obj_addr.ConfigObjectAddr):
        pass
    class ConfigObjectServ(config_obj_serv.ConfigObjectServ):
        pass
    class ConfigObjectNetServ(config_obj_net_serv.ConfigObjectNetServ):
        pass
    class ConfigObjectIndustrialAppServ(config_obj_indust_app_serv.ConfigObjectIndustrialAppServ):
        pass
    class ConfigNat(config_nat.ConfigNat):
        pass
    class ConfigDhcp(config_dhcp.ConfigDhcp):
        pass
    class ConfigOspf(config_ospf.ConfigOspf):
        pass
    class ConfigVrrp(config_vrrp.ConfigVrrp):
        pass
    class ConfigRstp(config_rstp.ConfigRstp):
        pass
    class ConfigRip(config_rip.ConfigRip):
        pass


from mdc.router.cli.rp_base.ng_router import config
from mdc.router.cli.rp_base.ng_router.cli import PROMPT_CONFIG

class ConfigBrg(config.Config):
    """
    Class for configuring Bridge interfaces.

        Args:
            session (object): A valid session object.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Attributes:
            _s (Session): A Session object used to communicate with the device.
            _args (list): A list of arguments passed to the function.
            _kwargs (dict): A dictionary of keyword arguments passed to the function.
    """

    def __init__(self, session: object, *args, **kwargs):
        super().__init__(session)
        self._args = args
        self._kwargs = kwargs

    def _set_default_prompts(self):
        """
        Set the default prompts.
        """
        self._s.set_default_prompts([PROMPT_CONFIG])

    def _getin(self):
        """
        Get into the interface lan.
        """
        self._s.command_expect(f'interface bridge')

    def getin(self):
        """
        Set the default prompts and get into the interface wan.

            Returns: 
                self: The instance of the object. 
        """
        self._set_default_prompts()
        self._getin()
        return self

    def set_brg_interface(self, name=None, ip=None, mask=None):
        """
        Set bridge interface parameters.

            Args:
                self (object): Instance of the class.
                name (str): Name of the bridge interface.
                ip (str): IP address of the bridge interface.
                mask (str): Subnet mask of the bridge interface.

            Returns:
                object: Instance of the class. 
        """
        if name:
            self._s.command_expect(f'name {name}')
        if ip and mask:
            self._s.command_expect(f'ip address {ip} {mask}')
        return self

    def disable_brg_interface(self):
        """
        Disable bridge interface.

            Returns: 
                object: Instance of the class. 
        """
        self._s.command_expect(f'shutdown')
        return self



from session import SshSession

class Hping3Service:
    """
    Class for controlling hping3 docker image over SSH session.

        Args:
            ssh_session (SshSession): SSH session object.
            sudo_paswd (str): sudo password.
        
         Attributes:
            _ssh (SshSession): An SSH session object.
    """

    def __init__(self, ssh_session: SshSession, sudo_paswd: str):
        self._ssh = ssh_session

    def _check_conneted(self):
        """
        Check if the environment is ready for running the tests.

        Raises:
            ValueError: If SSH session is not connected or docker image "sflow/hping3" is not found.
        """
        if not self._ssh.is_connected():
            raise ValueError('SSH session not connected')

    def check_env(self):
        """
        Check the environment for the current SSH connection.

        Raises:
            ValueError: If the docker image "sflow/hping3" is not found.
        """
        self._check_conneted()
        output = self._ssh.command('docker images')
        if 'sflow/hping3' not in output:
            raise ValueError('docker image "sflow/hping3" not found')

    # pylint: disable=too-many-arguments
    def send(self,
             host: str,
             tcpudp_flags: list,
             count: int,
             interval: str = 'faster',
             port: int = None):
        """ 
        Send attcks hping as attack
            Args:
                host (str): destination host address

                port (int): destination host's port

                interval (str): it can be fast, faster, flood or uX for X microseconds

                tcpudp_flag (list): it can be fin, syn, rst, push, ack or urg

                package_count (int): packet count
            
            Raises 
                ValueError: If it cannot make command from tcp/udp flag
        """
        self._check_conneted()

        flag_cmd_map = {
            'null': '',
            'fin': '-F',
            'syn': '-S',
            'rst': '-R',
            'push': '-P',
            'ack': '-A',
            'urg': '-U',
            'tcp': '-S',
            'udp': '-2',
            'icmp': '-1',
        }
        try:
            _flags = ' '.join([flag_cmd_map[f] for f in tcpudp_flags])
        except KeyError as e:
            raise ValueError('cannot make command from tcp/udp flag') from e

        _interval = {
            'fast': '--fast',
            'faster': '--faster',
            'flood': '--flood',
        }.get(interval, f'-i u{interval}')

        # example: docker run --rm --net=host sflow/hping3 192.168.127.31
        #           -R -S -i u10000 -p 80 -c 200
        # cmd = f'docker run --rm --net=host sflow/hping3 {host} -p {port} ' \
        #       f'{_flags} {_interval} -c {count}'
        cmd = f'docker run --rm --net=host sflow/hping3 {host} '
        cmd += '' if port is None else f'-p {port} '
        cmd += f'{_flags} '
        cmd += f'-c {count} '
        cmd += f'{_interval} '

        print(cmd)
        self._ssh.command(cmd)



import time

from mdc import Router
from abc import ABC, abstractmethod
from atb.const import get_control_class
from session import SshSession, TelnetSession, SerialSession
from utils.error_detector import cli_error_detector


class HostAbstract(ABC):

    @abstractmethod
    def login(self):
        pass

    @abstractmethod
    def logout(self):
        pass

    @abstractmethod
    def check_env(self):
        pass


class AtbHost(HostAbstract):

    def __init__(self, session: object, credential: dict, controls: list):
        self._s = session
        self._cred = credential
        self._ctrls = {}
        self._init_controls(controls)

    def _init_controls(self, controls):
        for ctrl in controls:
            ctrl_obj = ControlFactory.create(ctrl, self._s, self._cred)
            print(f'**************ctrl_obj: {ctrl_obj}**************')
            setattr(self, ctrl['name'], ctrl_obj)
            self._ctrls[ctrl['name']] = ctrl_obj

    def login(self):
        self._s.login(self._cred['username'], self._cred['password'])

    def logout(self):
        self._s.logout()

    def check_env(self):
        for _, obj in self._ctrls.items():
            print(f'**************check service: {_}**************')
            obj.check_env()

class SessionFactory:

    @staticmethod
    def create_ssh(conf: dict, auto_prompt_reset: bool = False):
        if conf['con_type'] == 'ssh':
            return SshSession(conf, auto_prompt_reset=auto_prompt_reset)
        else:
            raise ValueError(f'session type not supported: {conf["type"]}')
        
    @staticmethod
    def create_telnet(conf: dict, debug_output: bool = False):
        if conf['con_type'] == 'telnet':
            session = TelnetSession(conf, cli_error_detector, debug_output)
            session.open()
            return session
        else:
            raise ValueError(f'session type not supported: {conf["type"]}')

    @staticmethod
    def create_serial(conf: dict, debug_output: bool = False):
        if conf['con_type'] == 'serial':
            session = SerialSession(conf, cli_error_detector, debug_output)
            session.open()
            return session
        else:
            raise ValueError(f'session type not supported: {conf["type"]}')


class ControlFactory:

    @staticmethod
    def create(ctrl: dict, session: object, cred: dict):
        try:
            _type = ctrl['type']
            ctrl_cls = get_control_class(_type)
        except KeyError as e:
            raise ValueError('bad control config') from e

        return ctrl_cls(session, cred['password'])


class HostFactory:

    @staticmethod
    def create_atb(conf: dict):
        if conf['session']['con_type'] == 'ssh':
            session = SessionFactory.create_ssh(conf['session'], auto_prompt_reset=True)
        elif conf['session']['con_type'] == 'telnet':
            raise ValueError(f'Its not supported yet: {conf["session"]["con_type"]}')
        
        host_obj = AtbHost(session, conf['session']['credential'],
                                 conf['controls'])
        return host_obj
    
    @staticmethod
    def create_mdc_rp(conf: dict, debug_output:bool = False):
        if conf['session']['con_type'] == 'ssh':
            session = SessionFactory.create_ssh(conf['session'])
            session.login(conf['credential']['username'], conf['credential']['password'])
            host_obj = Router(session, conf['model'])
        elif conf['session']['con_type'] == 'telnet':
            session = SessionFactory.create_telnet(conf['session'], debug_output)
            session.login(conf['credential'])
            host_obj = Router(session, conf['model'])
        elif conf['session']['con_type'] == 'serial':
            session = SessionFactory.create_serial(conf['session'], debug_output)
            session.login(conf['credential'])
            host_obj = Router(session, conf['model'])
        
        return host_obj


from session import SshSession
from utils.decorator import root_access_required

class BirdService:
    """
    A class to manage bird server.

        Attributes:
            _ssh (SshSession): An instance of SshSession.
            _sudo_pwd (str): Password for sudo user.
    """

    def __init__(self, ssh_session: SshSession, sudo_paswd: str):
        self._ssh = ssh_session
        self._dir = "$HOME/bird"
        self._sudo_pwd = sudo_paswd

    def _check_conneted(self):
        """
        Check if the environment is valid.

            Raises:
                ValueError: If the SSH session is not connected or if no docker image cincan/tshark is found.
        """
        if not self._ssh.is_connected():
            raise ValueError('SSH session not connected')
        
    def check_env(self):
        self._check_conneted()
        output = self._ssh.command('docker images')
        if 'pierky/bird' not in output:
            self._ssh.command('docker pull pierky/bird:1.6.3', 
                              exact_prompts=['Downloaded newer image for pierky/bird:1.6.3'], 
                              timeout=120)
        output = self._ssh.command('ls')
        if 'bird' not in output:
            self._ssh.command(f'mkdir {self._dir}')
        self.clean_up()

    def start(self):
        docker_cmd = f'docker run --name bird --privileged --rm -dit --network=host \
            -v $HOME/bird:/etc/bird pierky/bird'
        self._ssh.command(docker_cmd)

    def clean_up(self):
        self._ssh.command("docker ps") 
        try:
            self._ssh.command("docker stop bird", 
                              exact_prompts=['bird'],
                              timeout=60)
        except:
            print("bird container is no actived")

    def set_ospf_config(self, 
                    router_id: str = "192.168.128.100", 
                    intf: str = "eth1", 
                    area_id: str = "0.0.0.0", 
                    priority: int = 1,
                    hello_interval: int = 10, 
                    dead_interval: int = 40):
        ospf_config= '''
protocol kernel {{
        scan time 60;
        import none;
        export all;   # Actually insert routes into the kernel routing table
}}

protocol device {{
        scan time 60;
}}

router id {};

protocol ospf {{
 export all;
 import all;
 area {} {{
  interface "{}" {{
   hello {};
   retransmit 5;
   cost 10;
   transmit delay 1;
   dead count 4;
   wait {};
   type broadcast;
   priority {};
   authentication none;
   #authentication cryptographic; password "ixnfo.com";
  }};
 }};
}}
'''.format(router_id, area_id, intf, hello_interval, dead_interval, priority)
        self._ssh.command(f"cat <<EOF > {self._dir}/bird.conf\n{ospf_config}\nEOF")



import threading
from threading import Thread


class ThreadWithResult(threading.Thread):
    """Extension of threading.Thread, with a `join` method which returns the thread's return value."""

    def __init__(self,
                 group=None,
                 target=None,
                 name=None,
                 args=(),
                 kwargs={},
                 *,
                 daemon=None):

        def function():
            self.result = target(*args, **kwargs)

        super().__init__(group=group, target=function, name=name, daemon=daemon)



import time
from behave import *
from steps import common, interface, login

@when(u'{action} igmp-snooping {version} with vlan {vid} on "{device}"')
def step_impl(context, action, version, vid, device):
    cfg = context.dut[device].go_config()
    if action == 'enable':
        cfg.IgmpSnooping(cfg.get_session()).set_igmp_snooping_vlan(action = action, 
                                                                vid = vid)
        cfg.IgmpSnooping(cfg.get_session()).set_igmp_snooping_querier(action = action, 
                                                                    vid = vid)
    elif action == 'disable':
        cfg.IgmpSnooping(cfg.get_session()).set_igmp_snooping_vlan(action = action, 
                                                                vid = vid)
    else:
        raise ValueError(f'input invalid action: {action}')


@when(u'set igmp-snooping query interval to {interval} on "{device}"')
def step_impl(context, interval, device):
    cfg = context.dut[device].go_config()
    cfg.IgmpSnooping(cfg.get_session()).set_igmp_snooping_interval(interval)
    

@when(u'send igmp {igmp_type} with following format from "{host}" to "{device}"')
def step_impl(context, host, device, igmp_type):
    igmp = {
        "query":{
            "sip": context.host_info[host]['testbed']['lan_host'],
            "dip": context.table[0]['dip'],
            "type": "IGMP(type=0x11)",
            "igmp_group_addr": context.table[0]['igmp_group_addr']
        },
        "join":{
            "sip": context.host_info[host]['testbed']['lan_host'],
            "dip": context.table[0]['dip'],
            "type": "IGMP(type=0x16)",
            "igmp_group_addr": context.table[0]['igmp_group_addr']
        },
        "leave":{
            "sip": context.host_info[host]['testbed']['lan_host'],
            "dip": context.table[0]['dip'],
            "type": "IGMP(type=0x17)",
            "igmp_group_addr": context.table[0]['igmp_group_addr']
        }
    }
    padding_n = "22"
    count = "1"
    iface = context.host_info[host]['nic']

    # fill igmp scapy script with args
    script = context.hosts[host].scapy.generate_igmp_scapy_script(igmp[igmp_type]["sip"], igmp[igmp_type]["dip"], 
                                                                padding_n, igmp[igmp_type]["type"], igmp[igmp_type]["igmp_group_addr"],
                                                                count, iface)
    from datetime import datetime
    import os

    # write the scapy script to file
    scapy_file_name = datetime.now().isoformat().replace(":", "_")[:19]
    with open(f'../lib/atb/service/scapy_template/{scapy_file_name}', "w") as script_file:
        script_content = script_file.write(script)

    # scp the file to remote host with scapy service
    os.system('sshpass -p "{0}" scp {1} {2}@{3}:~/'.format(context.host_info[host]['session']['credential']['password'],
                                                            f"../lib/atb/service/scapy_template/{scapy_file_name}",
                                                            context.host_info[host]['session']['credential']['username'],
                                                            context.host_info[host]['session']['host']))
    context.hosts[host].scapy.do_script(scapy_file_name)
    # Clean up
    os.system(f'rm ../lib/atb/service/scapy_template/{scapy_file_name}')
    

@then(u'igmp group should be {action} on "{device}"')
def step_impl(context, action, device):
    time.sleep(5)
    ret = context.dut[device].main().show_ip_igmp()
    flag = False
    for item in ret:
        flag = True if context.table[0]['igmp_group_addr'] == item['group'].replace(' ', '') else False
        if flag: break
    if action == 'added':
        assert flag, \
        f'\nexpect: {context.table[0]["igmp_group_addr"]}\nactual: {ret}'
    elif action == 'removed':
        assert not flag, \
        f'\nexpect: {context.table[0]["igmp_group_addr"]}\nactual: {ret}'
    else:
        raise ValueError(f'input wrong action: {action}')
    

@then(u'querier will be changed from "{device_1}" to "{device_2}"')
def step_impl(context, device_1, device_2):
    # skip the initial query packet
    time.sleep(5)
    if "DUT" in device_2:
        context.hosts[device_1].tshark.start(
            interface=context.host_info[device_1]["nic"],
            capture_filter='igmp')
        time.sleep(60)
        context.hosts[device_1].tshark.stop()
        capture_pkt = context.hosts[device_1].tshark.retrieve()
        querier_mac = f'{context.dut_info[device_2]["testbed"]["lan_ip"]} → 224.0.0.1'
        assert capture_pkt.count(querier_mac) >= 3, \
            f'querier does not change to {device_2}:\ncapture_pkt: {capture_pkt}'

    elif "DUT" in device_1:
        capture_pkt = context.hosts[device_2].tshark.start(
            interface=context.host_info[device_2]["nic"],
            capture_filter='igmp')
        time.sleep(60)
        context.hosts[device_2].tshark.stop()
        capture_pkt = context.hosts[device_2].tshark.retrieve()
        querier_mac = f'{context.dut_info[device_1]["testbed"]["lan_ip"]} → 224.0.0.1'
        assert capture_pkt.count(querier_mac) == 0, \
            f'querier does not change to {device_2}\ncapture_pkt: {capture_pkt}'
    else:
        raise ValueError(f'its only supported between HOST and DUT')

@when(u'wait for active time')
def step_impl(context):
    query_interval = context.table[0]["query_interval"]
    active_time = 2 * int(query_interval) + 10
    time.sleep(active_time)


@when(u'{action} global igmp-snooping on "{device}"')
def step_impl(context, action, device):
    cfg = context.dut[device].go_config()
    if action == 'enable':
        cfg.IgmpSnooping(cfg.get_session()).set_global_igmp_snooping(action)
    elif action == 'disable':
        cfg.IgmpSnooping(cfg.get_session()).set_global_igmp_snooping(action)
    else:
        raise ValueError(f'input invalid action: {action}')


from mdc.router.cli import common_func
from mdc.router.cli.rp_parser import parser
from mdc.router.cli.rp_base.ng_router import cli
from mdc.router.cli.rp_base.ng_router.cli import PROMPT_MAIN

class Main(cli.Cli):
    """
    Class for operating command at main prompt

        Args:
            session (obj): Session object.
            model (str): Model of the device.

        Attributes:
            _s (Session): A Session object used to communicate with the device.
            _model (str): Model of the device.
    """

    def __init__(self, session: object, model: str):
        super().__init__(session)
        self._model = model
        self._set_default_prompts()
        self._back_to_main()

    def _set_default_prompts(self):
        """
        Set default prompts for the session.
        """
        self._s.set_default_prompts([PROMPT_MAIN])

    def exit(self):
        """
        Exit from the current session.
        """
        self._s.command_expect('exit')

    def ping(self, host: str):
        """
        Ping a given host from the device.

            Args: 
                host (str): Host to be pinged.

            Returns: 
                str: Output of ping command. 
        """
        return parser.ping(self._s, host, self._model)

    def show_version(self):
        """
        Returns the version of the device.
            
            Returns:
                str: The version of the device.
        """
        return parser.show_version(self._s, self._model)

    def show_package(self):
        """
        Returns the package installed on the device.

            Returns:
                str: The package installed on the device. 
        """
        return parser.show_package(self._s, self._model)

    def show_system(self):
        """
        Returns information about system configuration on the device.

            Returns: 
                str: Information about system configuration on the device. 
        """
        return parser.show_system(self._s, self._model)

    def show_interface(self, name: str = 'lan'):
        """
        Returns information about interface configuration on the device for given name or lan by default if no name is provided .

            Args: 
                name (str, optional): Name of interface to get information about it's configuration . Defaults to 'lan'. 

            Returns:
                str: Information about interface configuration on the device for given name or lan by default if no name is provided .   #name is optional parameter with default value 'lan'  
        """
        return parser.show_interface(self._s, name, self._model)

    def show_vlan(self):
        """
        Show VLAN information on the device.

            Returns:
                dict: A dictionary containing the VLAN information.
        """
        return parser.show_vlan(self._s, self._model)

    def show_clock(self):
        """
        Show clock information on the device.

            Returns:
                dict: A dictionary containing the clock information. 
        """
        return parser.show_clock(self._s, self._model)

    def reload_factory_default(self):
        """
        Reloads the device to factory default settings.
        """
        common_func.reload_factory_default(self._s)

    def reload_factory_default_no_cert(self):
        common_func.reload_factory_default_no_cert(self._s)
        
    def show_dos(self):
        """
        Shows the Denial of Service (DoS) settings on the device.

            Returns: 
                str: The DoS settings on the device. 
        """
        return parser.show_dos(self._s, self._model)

    def show_ip_route(self):
        """
        Shows the IP routing table on the device.

            Returns: 
                dict: The IP routing table on the device.  
        """
        return parser.show_ip_route(self._s, self._model)

    def show_ip_igmp(self):
        """
        Shows the IP IGMP table on the device.

            Returns: 
                dict: The IP IGMP table on the device.  
        """
        return parser.show_ip_igmp(self._s, self._model)

    def show_logging_event_log_dos(self):
        """
        Shows logging event log for DoS attacks on the device.

            Returns:  
                dict: Logging event log for DoS attacks on the device.  
        """
        return parser.show_logging_event_log_dos(self._s, self._model)

    def show_logging_event_log_l3l7(self):
        """
        Retrieve the L3/L7 event log from the device.

            Returns:
                dict: L3/L7 event log.
        """
        return parser.show_logging_event_log_l3l7(self._s, self._model)

    def show_logging_event_log_system(self):
        """
        Retrieve the system event log from the device.

            Returns:
                dict: System event log. 
        """
        return parser.show_logging_event_log_system(
            self._s, self._model)

    def show_logging_event_log_malformed(self):
        """
        Retrieve the malformed event log from the device.

            Returns:
                dict: malformed event log.
        """
        return parser.show_logging_event_log_malformed(
            self._s, self._model)

    def show_l37_policy(self):
        """
        Retrieve the L3/L7 policy from the device.

            Returns: 
                dict: L3/L7 policy. 
        """
        return parser.show_l37_policy(self._s, self._model)

    def show_user_accounts(self):
        """
        Retrieve user accounts from the device. 

            Returns: 
                dict: User accounts and their settings. 
        """
        return parser.show_user_accounts(self._s, self._model)

    def show_syslog_setting(self):
        """
        Show syslog setting on the device. 

            Returns: None, command is executed on device without parsing output for now. 

        """
        self.command('show logging')

    # no parsing required for now
    def show_arp(self):
        """
        Execute the 'show arp' command on the device.

            Returns:
                str: Output of 'show arp' command.
        """
        return parser.show_arp(self._s, self._model)

    def show_mac_address_table(self):
        """
        Execute the 'show mac-address-table' command on the device.

            Returns:
                str: Output of 'show mac-address-table' command.
        """
        return parser.show_mac_address_table(self._s, self._model)

    def show_trusted_access(self):
        """
        Show trusted access information on the device.

            Returns:
                str: Output of 'show trusted access' command. 
        """
        return parser.show_trusted_access(self._s, self._model)

    def show_vrrp(self):
        """
         Show VRRP in human readable form. This is useful for debugging purposes.
         
         
            Returns: 
                A string of RLP formatted to be displayed on screen
        """
        return parser.show_vrrp(self._s, self._model)

    def show_rstp(self):
        """
         Show rstp global information in human readable form. This is useful for debugging purposes.
         
         
            Returns: 
                A string of RLP formatted to be displayed on screen
        """
        return parser.show_rstp(self._s, self._model)
    
    def show_rstp_port(self):
        """
         Show rstp port status in human readable form. This is useful for debugging purposes.
         
         
            Returns: 
                A string of RLP formatted to be displayed on screen
        """
        return parser.show_rstp_port(self._s, self._model)
    
    def show_ip_ospf_interface(self):
        """
         Show ospf interface in human readable form. This is useful for debugging purposes.
         
         
            Returns: 
                A string of RLP formatted to be displayed on screen
        """
        return parser.show_ip_ospf_interface(self._s, self._model)
    
    def show_ip_ospf_neighbor(self):
        """
         Show ospf neighbor in human readable form. This is useful for debugging purposes.
         
         
            Returns: 
                A string of RLP formatted to be displayed on screen
        """
        return parser.show_ip_ospf_neighbor(self._s, self._model)
    
    def show_ip_ospf_database(self):
        """
         Show ospf database in human readable form. This is useful for debugging purposes.
         
         
            Returns: 
                A string of RLP formatted to be displayed on screen
        """
        return parser.show_ip_ospf_database(self._s, self._model)

    def reload(self):
        """
        Reload the device with current configuration.
        """
        common_func.reload(self._s)

    def login(self):
        """
        Login to the device with credentials provided in constructor.
        """
        common_func.login(self._s)

    def clear(self,
                event_log: str = None,
                security_notification: str = None):
        """
        Clear the event log or security notification status. 
            Args: 
                event_log (str): The type of event log to clear. Valid values are: "all", "vpn", "system", "trusted-access", "malformed", "dos", "l3l7-policy", "dpi", "adp", "ips" and "session-control". 
                security_notification (str): The type of security notification to clear. 
            Raises: 
                ValueError: If the input event log type is invalid. 
        """
        log_type = [
            "all", "vpn", "system", "trusted-access", "malformed", "dos",
            "l3l7-policy", "dpi", "adp", "ips", "session-control"
        ]
        if event_log != None:
            if event_log in log_type:
                if event_log == "all":
                    self._s.command_expect('clear logging event-log')
                else:
                    self._s.command_expect(
                        f'clear logging event-log {event_log}')
            else:
                raise ValueError(
                    f'input invalid event log type: {event_log}')
        elif security_notification != None:
            self._s.command_expect('clear security-notification status')

    def set_backdoor(self):
        """
        Set up a backdoor for the device.  
            Raises: 
                Exception: If failed to setup backdoor.
        """
        self._s.sendcontrol('t')
        cmd_args = [
            ('ieisecureedr moxaiei89191230', [r'/ #']),
            ('passwd -ud root', [r'/ #']),
            # ('exit', [PROMPT_MAIN]),
            # ('exit', [PROMPT_MAIN]),
        ]
        for args in cmd_args:
            r = self._s.command_expect(args[0], prompts=args[1])
            if not r['matched']:
                print(f"return = {r}, command = {args[0]}")
                raise Exception("failed to setup backdoor")
        return self

    def clear_ssh_telnet_limit(self):
        """
        This step may require modification for future usage.
        Due to the DUT's limit on the number of SSH and Telnet sessions, 
        we need to clear the limitation before logging in and out multiple times.
        """
        self.set_backdoor()
        self._s.command_expect('iptables -F PortScan')
        self._s.command_expect('exit')
        # disable ctrl-t mode for serial control
        self._s.sendcontrol('t')

    def import_config(self,
                        tftp_ip: str = None,
                        usb: bool = False,
                        file_name: str = "test.ini"):
        """
        Import a configuration file from a TFTP server or USB.

            Args:
                tftp_ip (str): The IP address of the TFTP server.
                file_name (str, optional): The name of the configuration file to import. Defaults to "test.ini".

            Returns:
                str: A string indicating whether the config file was imported successfully or not. 
        """
        if tftp_ip != None and usb == False:
            if file_name == 'configuration file':
                return self._s.command_expect(
                    f'copy tftp {tftp_ip} config-file test.ini',
                    prompts=['Config file import successfully'],
                    timeout=25)
            else:
                return self._s.command_expect(
                    f'copy tftp {tftp_ip} config-file {file_name}',
                    prompts=['Config file import successfully'],
                    timeout=25)
        elif tftp_ip == None and usb == True:
            if file_name == 'configuration file':
                return self._s.command_expect(
                    f'copy usb test.ini',
                    prompts=['Config file import successfully'],
                    timeout=25)
            else:
                return self._s.command_expect(
                    f'copy usb {file_name}',
                    prompts=['Config file import successfully'],
                    timeout=25)

    def export_config(self, tftp_ip: str, file_name: str = "test.ini"):
        """
        Export the running configuration to a TFTP server.

            Args:
                tftp_ip (str): The IP address of the TFTP server.
                file_name (str, optional): The name of the configuration file to export. Defaults to "test.ini".

            Returns:
                str: A string indicating whether the config was uploaded successfully or not. 
        """
        return self._s.command_expect(
            f'copy running-config tftp {tftp_ip} {file_name}',
            prompts=['Configuration Upload Success!'], timeout=30)

    def export_eventlog(self, server: str="tftp", server_ip: str = None, category: str= "system"):
        """
         Export event log to file. This will create a file in the current directory with the contents of the event log.
         
            Args:
                server: Name of the server to export the event log to. Default : tftp. You can specify a different server by using the server_ip parameter.
                server_ip: IP address of the server to export the event log to.
                category: The category of the event log. Valid values are system vpn trust_access malformed_packets dos_policy device_lockdown L3L7_
        """
        event_map = {
            "system": "0",
            "vpn": "1",
            "trust_access": "2",
            "malformed_packets": "3",
            "dos_policy": "4",
            "device_lockdown": "5",
            "l3l7_policy": "6",
            "protocol_filter_policy": "7",
            "adp": "8",
            "ips": "9",
            "session_control": "10",
            "l2_policy": "11"
        }[category]
        server_map = {
            "tftp": "1", 
            "usb": "2", 
            "scp": "3", 
            "sftp": "4"
        }[server]
        return self._s.command_expect(
            f'copy event-log {event_map} {server_map} {server_ip}',
            prompts=['Event Log File Exporting is Complete.'])

    def get_running_config(self, con_type='telnet'):
        """
        Retrieve and return the running configuration from a device instance.

            Args:
                con_type (str): The connection type of DUT. Default to 'telnet'.

            Returns: 
                str: A string containing the running configuration of the device instance.
        """
        self.command('terminal length 0')
        return self._s.command_expect('show running-config',
                                        prompts=[PROMPT_MAIN],
                                        exact_str=True,
                                        timeout=5)['data']

    def upgrade_firmware(
            self,
            tftp_ip: str,
            file_name: str = "FWR_EDR-G9010_V3.0_Build_22122312.rom "):
        """
        Upgrade the firmware of the device from a given tftp server.

            Args:
                tftp_ip (str): The IP address of the TFTP server.
                file_name (str): The name of the firmware file to be upgraded. Defaults to "FWR_EDR-G9010_V3.0_Build_22122312.rom".

            Returns:
                dict: A dictionary containing the response from the device after upgrading the firmware. 
        """
        ret = self._s.command_expect(
            f'copy tftp {tftp_ip} device-firmware {file_name}',
            prompts=['login', 'Y/N'],
            timeout=180)
        # print('upgrade fw ret: ', ret)
        if ret['pattern'] == 'Y/N':
            # print('ret pattern == Y/N: ', ret)
            return self._s.command_expect('Y',
                                            prompts=['login'],
                                            exact_str=True,
                                            timeout=180)
        return ret

    def upgrade_package(
            self,
            tftp_ip: str,
            pkg_type: str,
            file_name: str = "Security_EDR-G9010_V5.5.4_Build_23020217.pkg"
    ):
        """
        Upgrade a package from a given tftp server.

            Args:
                tftp_ip (str): The IP address of the TFTP server. 
                pkg type (str): The type of package to upgrade e.g security or application etc.. 
                file name (str): The name of the package file to be upgraded e.g Security EDR-G9010 V5.5 Build 23020217 etc.. Defaults to "Security EDR-G9010 V5 Build 23020217".

            Returns: 
                dict: A dictionary containing the response from the device after upgrading the package 
        """
        ret = self._s.command_expect(
            f'package upgrade {pkg_type} tftp {tftp_ip} {file_name}',
            timeout=180)
        return ret

    def save(self):
        """
        Save changes made to device configuration

            Returns: 
                dict: A dictionary containing response from device after saving changes
        """
        return self._s.command_expect('save',
                                        prompts=[PROMPT_MAIN],
                                        exact_str=True,
                                        timeout=6)

    def set_config_file(self,
                        dig_signature: str,
                        data_encrypt: str = None,
                        encrypt_pwd: str = None):
        """
        Set the configuration file.

            Args:
                dig_signature (str): Configuration File Signature. Should be either "enable" or "disable".
                data_encrypt (str, optional): Signature Information. Should be either "sensitive" or "all".
                encrypt_pwd (str, optional): Encryption password.
                
            Raises:
                ValueError: If an invalid dig_signature or data_encrypt is given as input. 
        """
        if dig_signature in ["enable", "disable"]:
            if dig_signature == "enable":
                self._s.command_expect("config-file digital-signature",
                                        prompts=[PROMPT_MAIN])
            else:
                self._s.command_expect("no config-file digital-signature",
                                        prompts=[PROMPT_MAIN])
        else:
            raise ValueError(
                f'input invalid dig_signature: {dig_signature}')

        if data_encrypt:
            if data_encrypt in ["sensitive", "all"]:
                self._s.command_expect(
                    f"config-file data-encryption {data_encrypt}",
                    prompts=[PROMPT_MAIN])
            else:
                raise ValueError(
                    f'input invalid data_encrypt: {data_encrypt}')

        if encrypt_pwd:
            self._s.command_expect(
                f"config-file encryption-password {encrypt_pwd}",
                prompts=[PROMPT_MAIN])

    def show_object(self):
        ret = parser.show_object(self._s, self._model)
        return ret

    def show_session_ctrl(self):
        ret = parser.show_session_ctrl(self._s, self._model)
        return ret

    def show_lldp_table(self):
        ret = parser.show_lldp_table(self._s, self._model)
        return ret
    
    def show_l2tp_setting(self):
        return parser.show_l2tp_setting(self._s, self._model)
    
    def show_snmp(self):
        return parser.show_snmp(self._s, self._model)



from mdc.router.cli.rp_base.lg_router.tn5000 import config
from mdc.router.cli.rp_base.lg_router.tn5000.cli import PROMPT_CONFIG

class ConfigObjectServ(config.Config):
    """
    Class for configuring User-defined Service Object.

        Args:
            session (object): The session object.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Attributes:
            _s (Session): A Session object used to communicate with the device.
            _args (list): A list of arguments passed to the function.
            _kwargs (dict): A dictionary of keyword arguments passed to the function.
    """

    def __init__(self, session: object, *args, **kwargs):
        super().__init__(session)
        self._args = args
        self._kwargs = kwargs

    def _set_default_prompts(self):
        """
        Set the default prompts.
        """
        self._s.set_default_prompts([PROMPT_CONFIG])

    # Reserve flexibility for function override
    def _getin(self):
        """
        Get into the config-obj-serv prompt.
        """
        # default apply the first element as parameter
        service = self._args[0] if len(self._args) != 0 else None
        if service is None:
            raise Exception(f"Input parameter is empty!")
        self._s.command_expect(f'object service {service}')

    def getin(self):
        """
        Set the default prompts and get into the config-obj-serv prompt.

            Returns: 
                self: The instance of the object. 
        """
        self._set_default_prompts()
        self._getin()
        return self

    def set_object_serv(self,
                        name: str,
                        port: str = None,
                        icmp_type: str = None,
                        icmp_code: str = None,
                        ipproto: str = None):
        """
        Create an object with the given name and IP address.

            Args:
                name (str, optional): The name of the object to create.
                port (str, optional): The port of the object to create.
                
            Returns:
                self: The instance of the object. 
        """
        if name:
            self._s.command_expect(f'name {name}')
        if port:
            self._s.command_expect(f'port {port}')
        if icmp_type:
            self._s.command_expect(f'icmp-type {icmp_type}')
        if icmp_code:
            self._s.command_expect(f'icmp-code {icmp_code}')
        if ipproto:
            self._s.command_expect(f'ipproto {ipproto}')

        # quit to save
        self._s.command_expect(f'exit')
        return self



from mdc.router.cli.rp_base.ng_router import config
from mdc.router.cli.rp_base.ng_router.cli import PROMPT_CONFIG

class ConfigL37Policy(config.Config):
    """
    Class for configuring L37 Policy.

        Args:
            session (object): A valid session object.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Attributes:
            _s (Session): A Session object used to communicate with the device.
            _args (list): A list of arguments passed to the function.
            _kwargs (dict): A dictionary of keyword arguments passed to the function.
    """

    def __init__(self, session: object, *args, **kwargs):
        super().__init__(session)
        self._args = args
        self._kwargs = kwargs

    def _set_default_prompts(self):
        """
        Set the default prompts.
        """
        self._s.set_default_prompts([PROMPT_CONFIG])

    def _getin(self):
        """
        Get into the interface lan.
        """
        self._s.command_expect(f'l3l7-policy')

    def getin(self):
        """
        Set the default prompts and get into the interface wan.

            Returns: 
                self: The instance of the object. 
        """
        self._set_default_prompts()
        self._getin()
        return self

    def set_l37_policy(self,
                        name=None,
                        enable=None,
                        disable=None,
                        logging=None,
                        itf=None,
                        policy_action=None,
                        mode=None,
                        sip=None,
                        dip=None,
                        sport=None,
                        dport=None,
                        smac=None,
                        dmac=None):
        """
        Set L37 policy on the device.

            Args:
                name (str): Name of the policy. 
                enable (bool): Enable the policy. 
                disable (str): Disable the policy. 
                logging (str): Enable or disable logging for this policy. 
                itf (dict): Interface in and out for the policy. 
                policy_action (str): Allow or deny action for this policy. 
                mode (str): Mode of this policy, ip, ip-mac or mac. 
                sip (str): Source IP address for this policy. 
                dip (str): Destination IP address for this policy. 
                sport (str): Source port for this policy. 
                dport (str): Destination port for this policy. 
                smac (str): Source MAC address for this policy. 
                dmac (str): Destination MAC address for this policy.  

            Returns:   
                self: The instance of the object.   
        """
        if itf["in"] == 'BRG':
            itf["in"] = 'BRG_LAN'
        if itf["out"] == 'BRG':
            itf["out"] = 'BRG_LAN'
        if name:
            self._s.command_expect(f'name {name}')
        if enable == True:
            self._s.command_expect('enable')
        if logging == "enable":
            self._s.command_expect('logging')  # left to be complete
        if itf:
            self._s.command_expect(f'interface {itf["in"]} {itf["out"]}')
        if policy_action in ["allow", "deny"]:
            self._s.command_expect(f"action {policy_action}")
        if mode in ["ip", "ip-mac", "mac"]:
            self._s.command_expect(f"mode {mode}")
        if sip:
            self._s.command_expect(f"src-ip {sip}")
        if dip:
            self._s.command_expect(f"dst-ip {dip}")
        if sport:
            self._s.command_expect(f"src-port {sport}")
        if dport:
            self._s.command_expect(f"dst-port {dport}")
        if smac:
            self._s.command_expect(f"src-mac {smac}")
        if dmac:
            self._s.command_expect(f"dst-mac {dmac}")
        if disable in ["no"]:
            pass  # left to be complete
        return self



import json
from datetime import datetime

def create_debug_output(option: bool, name_prefix: str, con_type: str,
                        host: str, port):
    """
    Creates debug output for a session. 

        Args: 
            option (bool or file object): If True, session information and output will be logged. If False, nothing will be logged. Or set a file object for writing console output only. 
            name_prefix (str): Prefix for the debug output files. 
            con_type (str): Type of connection used in the session. 
            host (str): Host address used in the session. 
            port (int): Port number used in the session. 
            
        Returns: 
            output_logfile (file object): File object of the logfile created if option is True or set to a file object; None otherwise.
    """
    time_str = datetime.now().isoformat()
    name = f'{name_prefix}_{time_str}'

    if option is True:
        # create debug info
        with open(f'{name}.info.json', 'w', encoding='utf8') as f:
            json.dump(
                {
                    'start_time': time_str,
                    'connection_type': con_type,
                    'host': host,
                    'port': port
                },
                f,
                indent=2)
        # create logfile object
        # pylint: disable=consider-using-with
        output_logfile = open(f'{name}.output.txt', 'w', encoding='utf8')
    elif option:
        output_logfile = option
    else:
        output_logfile = None
    return output_logfile



from ixnetwork_restpy import *
from ixnetwork_restpy.errors import IxNetworkError
import time, json, sys, os, re, datetime, traceback

class ixia:
	def __init__(self, ixia_ip, username, password, session_name=None, log_level='all', log_file_name='ixia.log'):
		self._s = self._connect_ixia(ixia_ip, username, password, session_name, log_level, log_file_name)
		self._ix_network = self._s.Ixnetwork
		self._vport = None
		self._debug = False
		self._timestamp = None

	def _connect_ixia(self, ixia_ip, username, password, session_name=None, log_level='all', log_file_name='ixia.log'):
		session = SessionAssistant(IpAddress=ixia_ip, RestPort=None, 
									UserName=username, Password=password, 
									SessionName=session_name, SessionId=None, ApiKey=None, 
									ClearConfig=True, LogLevel=log_level, 
									LogFilename=log_file_name)
		return session

	def _set_timestamp(self):
		self._timestamp = datetime.datetime.now().strftime('%Y%m%d-%H%M')

	def _add_timestamp_to_file(self, file_name):
		if self._timestamp != None:
			file_name = file_name.split('/')[-1]
			file_name_with_timestamp = '{}_{}.{}'.format(file_name.split('.')[0], self._timestamp, file_name.split('.')[1])
			return file_name_with_timestamp

	def _add_prepended_to_file(self, prepended, file_name):
		file_name = file_name.split('/')[-1]
		file_name_with_prepended = '{}_{}'.format(prepended, file_name)
		return file_name_with_prepended

	def _get_qt_type(self, qt_handle):
		match = re.search('/api/.*/quickTest/(.*)/[0-9]+', qt_handle.href)
		qt_type = match.group(1)
		return qt_type

	def _check_qt_initialization(self, qt_handle):
		success_status = ['TransmittingComplete', 'TransmittingFrames', 'WaitingForStats', 'CollectingStats', 'TestEnded']
		qt_apply_states = ['InitializingTest', 'ApplyFlowGroups', 'SetupStatisticsCollection']
		apply_qt_counter = 120

		for timer in range(1, 31):
			current_action = self._get_qt_current_action(qt_handle)
			self._ix_network.info('\n\nVerify QuickTest Initialization - Current Action: {}\n'.format(current_action))
			if current_action == 'TestEnded':
				raise Exception('\n\nVerify QuickTest Initialization - QuickTest failed during initialization: {}'.format(qt_handle.Results.Status))

			if timer < 30 and current_action == 'None':
				self._ix_network.info('\n\nVerify QuickTest Initialization - Current State = %s\n\tWaiting %s/30 seconds to change state\n' % (current_action, timer))
				time.sleep(1)
				continue
			else:
				break

			if timer == 20 and current_action == 'None':
				raise Exception('\n\nQuick Test is stuck.')

		for counter in range(1, apply_qt_counter+1):
			current_action = self._get_qt_current_action(qt_handle)
			self._ix_network.info('\n\nVerify QuickTest Initialization - Current State: %s  Expecting: TransmittingFrames\n\tWaiting %s/%s seconds\n' % (current_action, counter, apply_qt_counter))

			if current_action == 'TestEnded':
				raise Exception('\n\nVerify QuickTest Initialization - QuickTest failed: {}'.format(qt_handle.Results.Status))
			elif current_action == None:
				currentAction = 'ApplyingAndInitializing'

			if counter < apply_qt_counter and current_action not in success_status:
				time.sleep(1)
				continue
			elif counter < apply_qt_counter and current_action in success_status:
				self._ix_network.info('\n\nVerify QuickTest Initialization is done applying configuration and has started transmitting frames\n')
				break  

			if counter == apply_qt_counter:
				if current_action not in success_status:
					if current_action == 'ApplyFlowGroups':
						self._ix_network.info('\n\nVerify QuickTest Initialization: IxNetwork is stuck on Applying Flow Groups. You need to go to the session to FORCE QUIT it.\n')

					raise Exception('\n\nVerify QuickTest Initialization: is stuck on %s. Waited %s/%s seconds' % (current_action, counter, apply_qt_counter))

	def _get_qt_current_action(self, qt_handle):
		timer = 10
		for counter in range(1, timer+1):
			current_actions = qt_handle.Results.CurrentActions

			self._ix_network.info('\n\nGet QuickTest Current Action:\n')
			for each_current_action in qt_handle.Results.CurrentActions:
				self._ix_network.info('\t{}'.format(each_current_action['arg2']))

			self._ix_network.info('\n')

			if counter < timer and current_actions == []:
				self._ix_network.info('\n\nGet QuickTest Current Action is empty. Waiting %s/%s\n\n' % (counter, timer))
				time.sleep(1)
				continue
			elif counter < timer and current_actions != []:
				break
			elif counter == timer and current_actions == []:
				raise Exception('\n\nGet QuickTest Current Action: Has no action')

		return current_actions[-1]['arg2']

	def _monitor_qt_running_progress(self, qt_handle, get_progress_interval=10, max_retries=10):
		flag_running_break = 0
		flag_traffic_started = 0
		counter_wait_for_running_progress = 0
		counter_connection_failure = 0
		counter = 1

		while True:
			flag_connect_to_api_server = False
			while True:
				try:					
					is_running = qt_handle.Results.IsRunning
					current_runnung_progress = qt_handle.Results.Progress
					self._ix_network.debug('\n\nMonitor QuickTest Running Progress - is_runing: {}'.format(is_running))
					break
				except:
					self._ix_network.debug('\n\nMonitor QuickTest Running Progress: Failed to connect to API server {}/{} times\n'.format(counter_connection_failure, max_retries))
					if counter_connection_failure == max_retries:		
						raise Exception('\n\nMonitor QuickTest Running Progress: Giving up trying to connecto the the API server after {} attempts\n'.format(max_retries))
					elif counter_connection_failure <= max_retries:						
						counter_connection_failure += 1
						time.sleep(3)
						continue

			self._ix_network.info('\n\nMonitor QuickTest Running Progress - is_running: {}  Current Running Progress: {}\n'.format(is_running, current_runnung_progress))

			if is_running == True:
				if bool(re.match('^Trial.*', current_runnung_progress)) == False:
					if counter_wait_for_running_progress < 40:						
						self._ix_network.info('\n\nMonitor QuickTest Running Progress: Waiting for trial runs {0}/30 seconds\n'.format(counter_wait_for_running_progress))
						counter_wait_for_running_progress += 1
						time.sleep(1)
					elif counter_wait_for_running_progress == 40:					
						raise Exception('\n\nMonitor QuickTest Running Progress: is_running=True. QT is running, but no quick test iteration stats showing after 40 seconds.')
				else:			
					flag_traffic_started = 1
					time.sleep(get_progress_interval)
					continue
			else:
				if flag_traffic_started == 1:					
					self._ix_network.info('\n\nMonitor QuickTest Running Progress: is_running=False. Quick Test ran and is complete\n\n')
					return True
				elif flag_traffic_started == 0 and flag_running_break < 40:				
					self._ix_network.info('\n\nMonitor QuickTest Running Progress: isRunning=False. QT did not run yet. Wait {0}/40 seconds\n\n'.format(flag_running_break))
					flag_running_break += 1
					time.sleep(1)
					continue
				elif flag_traffic_started == 0 and flag_running_break == 40:					
					raise Exception('\n\nMonitor QuickTest Running Progress: Quick Test failed to start:: {}'.format(qt_handle.Results.Status))

	def _save_qt_csv_files(self, qt_handle, path, csv_file='all', prepended=None, include_timestamp=False):
		results_path = qt_handle.Results.ResultPath
		report = self._s.Session.GetFileList(remote_directory=results_path)

		self._ix_network.info('\n\nGet QuickTest Csv Files: %s\n' % results_path)

		if csv_file == 'all':
			csv_file_list = ['AggregateResults.csv', 'PortInfo.csv', 'results.csv', 'logFile.txt', 'PortMap.csv', 'info.csv']
		else:
			if type(csv_file) is not list:
				csv_file_list = [csv_file]
			else:
				csv_file_list = csv_file

		for each_file in csv_file_list:
			source = results_path+'/{0}'.format(each_file)
			try:
				self._ix_network.info('\n\nCopying file from Linux API server:{} to local Linux: {}\n'.format(source, each_file))
				self._copy_api_server_file_to_local_linux(source, path, prepended=prepended, include_timestamp=include_timestamp)
			except Exception as errMsg:
				self._ix_network.warn('\n\nCopying file from API server ERROR: {}\n'.format(errMsg))

	def _copy_api_server_file_to_local_linux(self, api_server_path_and_file_name, local_path, prepended=None, include_timestamp=False):
		file_name = api_server_path_and_file_name.split('/')[-1]
		file_name = file_name.replace(' ', '_')
		if prepended != None:
			file_name = self._add_prepended_to_file(prepended, file_name)
		if include_timestamp:
			file_name = self._add_timestamp_to_file(file_name)
		dest_path = '{}/{}'.format(local_path, file_name)
		self._ix_network.info('\nCopying file from API server:{} -> {}'.format(api_server_path_and_file_name, dest_path))
		self._s.Session.DownloadFile(api_server_path_and_file_name, dest_path)

	def _verify_ngpf_is_layer3(self):
		num = self._ix_network.Topology.find().index+2
		res = True
		for i in range(1, num):
			if self._ix_network.Topology.find(f'Topology {i}').DeviceGroup.find().Ethernet.find().Ipv4.find():
				print('\n\nTopology is Layer3: {}\n'.format(self._ix_network.Topology.find(f'Topology {i}').DeviceGroup.find().Ethernet.find().Ipv4.find().href))
			elif self._ix_network.Topology.find(f'Topology {i}').DeviceGroup.find().Ethernet.find().Ipv6.find():
				print('\n\nTopology is Layer3: {}\n'.format(self._ix_network.Topology.find(f'Topology {i}').DeviceGroup.find().Ethernet.find().Ipv6.find().href))
			else:
				print('\n\nTopology is Layer3: False\n')
				res = False
		return res

	def start_all_protocol(self):
		verify_ngpf_is_layer3 = self._verify_ngpf_is_layer3()
		if verify_ngpf_is_layer3 == True:
			self._ix_network.StartAllProtocols(Arg1='sync')
			self._ix_network.info('Verify protocol sessions\n')
			protocol_summary = self._s.StatViewAssistant('Protocols Summary')
			protocol_summary.CheckCondition('Sessions Not Started', protocol_summary.EQUAL, 0)
			protocol_summary.CheckCondition('Sessions Down', protocol_summary.EQUAL, 0)
			self._ix_network.info(protocol_summary)
		else:
			self._ix_network.info('Topology is not Layer 3: no need to start protocol\n')

	def remove_session(self):
		if not self._debug:
			self._s.Session.remove()

	def reserve_port(self, port_list, ixia_ip, force_take_port_owner_ship=True):
		self._ix_network.info('Assign ports...')
		port_map = self._s.PortMapAssistant()
		vport = dict()
		for index, port in enumerate(port_list):
			port_name = 'port_{}'.format(index+1)
			port_name = self._ix_network.Vport.find()[index].Name
			card_id = port.split('/')[0]
			port_id = port.split('/')[1]
			vport[port_name] = port_map.Map(IpAddress=ixia_ip,
											CardId=card_id,
											PortId=port_id,
											Name=port_name)
		port_map.Connect(force_take_port_owner_ship)

		self._vport = vport

	def set_media_type(self, port_media):
		for vport in self._ix_network.Vport.find():
			if vport.Type == 'novusTenGigLan':
				vport.L1Config.NovusTenGigLan.Media = port_media

	def load_ixia_cfg(self, config_file):
		self._ix_network.LoadConfig(Files(config_file, local_file=True))

	def run_rfc_2544(self, report_path, test_model):
		self._set_timestamp()
		prepended = test_model + '_rfc-2544'
		report_path = report_path + '_' + self._timestamp
		print('----------report_path----------')
		print(report_path)
		print('----------report_path----------')
		os.mkdir(report_path)
		qt_handle = self._ix_network.QuickTest.Rfc2544throughput.find()[0]
		self._ix_network.info('\n\nQuickTest Handle: {}\n'.format(qt_handle))
		qt_type = self._get_qt_type(qt_handle)
		test_title = '{} - {}'.format(qt_type, qt_handle.Name)
		self._ix_network.info('\n\nExecuting QuickTest: {}\n'.format(test_title))
		qt_handle.Apply()
		qt_handle.Start()
		self._check_qt_initialization(qt_handle)
		self._monitor_qt_running_progress(qt_handle)
		self._save_qt_csv_files(qt_handle, path=report_path, prepended=prepended, include_timestamp=True)




from behave import *
import time
from datetime import datetime
import os

@when('{action} broadcast forwarding on "{device}"')
def step_impl(context, action, device):
    context.dut[device].go_config().set_broadcast_forward_status(action = action)


@when('add broadcast forwarding rule on "{device}"')
def step_impl(context, device):
    context.dut[device].go_config().add_broadcast_forward_rule(context.table[0]['In_interface'],
                                                                context.table[0]['Out_interface'],
                                                                context.table[0]['UDP_Port'])
    context.dut[device].main().save()


@when('"{host}" send broadcast packet to its subnet')
def step_impl(context, host):
    sip = context.host_info[host]['testbed']['cur_host']
    ip_parts = sip.split(".")
    ip_parts[3] = "255"
    subnet_ip = ".".join(ip_parts) 
    dip =  subnet_ip
    dport = context.table[0]['dest_port']
    iface = context.host_info[host]['nic']
    script = context.hosts[host].scapy.generate_broadcast_scapy_script(sip, dip, dport, iface, 1)

    # write the scapy script to file
    scapy_file_name = datetime.now().isoformat().replace(":", "_")[:19]
    with open(f'../lib/atb/service/scapy_template/{scapy_file_name}', "w") as script_file:
        script_content = script_file.write(script)

    # scp the file to remote host with scapy service
    os.system('sshpass -p "{0}" scp {1} {2}@{3}:~/'.format(context.host_info[host]['session']['credential']['password'],
                                                            f"../lib/atb/service/scapy_template/{scapy_file_name}",
                                                            context.host_info[host]['session']['credential']['username'],
                                                            context.host_info[host]['session']['host']))
    
    context.hosts[host].scapy.do_script(scapy_file_name)
    # Clean up
    os.system(f'rm ../lib/atb/service/scapy_template/{scapy_file_name}')
    time.sleep(8)

    context.broadcast_forward = {
        "sip":sip,
        "dip":dip
    }


@then('"{host_recv}" {result} receive broadcast packet from "{host_send}"')
def step_impl(context, host_recv, result, host_send):
    """
    result can be "shall" or "shall not"
    """
    context.hosts[host_recv].tshark.stop()
    ret = context.hosts[host_recv].tshark.retrieve()
    
    if len(ret) == 0:
        raise ValueError('packet receiving fails')

    pkt_count = 0

    ret_lines = ret.split('\n')
    for a in ret_lines:
        arr = a.lstrip(' ').split(' ')
        # skip if not tshark packet
        if arr[0].isdigit() == False:
            continue
        if arr[3] != '→':
            raise ValueError(f'unexpected tshark format {a}')

        # Match Source and Destination IP Address
        if arr[2] == context.broadcast_forward['sip'] and arr[4] == context.broadcast_forward['dip']:
            pkt_count += 1

    if result == 'shall not':
        assert pkt_count == 0, f'{host_recv} expect to receive no packets but receive {pkt_count}'
    elif result == 'shall':
        assert pkt_count > 0, f'{host_recv} expect to receive packets but receive {pkt_count}'


import os
import time
from behave import *
from datetime import datetime
from steps import common, interface, login


@given(u'modify RADIUS config on "{host}"')
def step_impl(context, host):
    context.hosts[host].radius.add_client(client_name='radiusclient')

    for row in context.table:
        raddb_username = row['username']
        raddb_password = row['password']
        raddb_service_type = row['service_type']
        context.hosts[host].radius.add_users(username=raddb_username,
                                             password=raddb_password,
                                             service_type=raddb_service_type)


@when(u'set Login Authentication as {auth_mode} on "{device}"')
def step_impl(context, device, auth_mode):
    auth_mode_map = {
        "Local":            "local",
        "RADIUS":           "radius",
        "RADIUS_Local":     "radius_local"    
    }[auth_mode]
    context.dut[device].go_config().set_authentication_login_mode(auth_mode=auth_mode_map)


@when(u'set authentication type as {auth_type} for the RADIUS server on "{device}"')
def step_impl(context, device, auth_type):
    context.dut[device].go_config().set_radius_server_auth_type(auth_type=auth_type)


@when(u'start Radius server on "{host}"')
def step_impl(context, host):
    context.hosts[host].radius.build_local_image()
    context.hosts[host].radius.start()


@when(u'stop Radius server on "{host}"')
def step_impl(context, host):
    context.hosts[host].radius.stop()


@then(u'"{host}" "{action}" to login "{device}" as "{username}" with "{password}" via "{console_type}" from {interface}')
@then(u'"{host}" "{action}" to login "{device}" as {username} with password {password} via "{console_type}"')
def step_impl(context, host, action, device, username, password, console_type, interface='lan'):
    target_ip = context.dut_info[device]["testbed"][f"{interface.lower()}_ip"]
    container_flag = False
    result_map = {}

    for row in context.table:
        success_action_flag = {'success': True, 'fail': False}[row['action']]
        result_map[row['console_type']] = False
        wait_time = row['wait_time'] if context.table and 'wait_time' in context.table.headings else 3
        ret = []

        if row['console_type'] == 'HTTP':
            ret_mark = ['access_token', 'login success']
            _filename = f'login_check_http_{datetime.now().isoformat()}'
            cmd = f"curl --location 'http://{target_ip}/api/v1/auth/login' "
            cmd += f"--header 'Content-Type: text/plain' "
            cmd += f"--data '{{\"username\": \"{username}\", \"password\": \"{password}\"}}' -k"
            cmd += f" > {_filename}.log"

            context.hosts[host].shellcmd._ssh.command(cmd, timeout=int(wait_time)).lower()
            for _ in range(int(wait_time) + 4):
                ret = context.hosts[host].shellcmd._ssh.command(f'cat {_filename}.log')
                if any(map(lambda s: s in ret, ret_mark)) or 'Invalid Authentication Information' in ret:
                    break
                time.sleep(1)

        elif row['console_type'] == 'HTTPS':
            ret_mark = ['access_token', 'login success']
            _filename = f'login_check_https_{datetime.now().isoformat()}'
            cmd = f"curl --location 'https://{target_ip}/api/v1/auth/login' "
            cmd += f"--header 'Content-Type: text/plain' "
            cmd += f"--data '{{\"username\": \"{username}\", \"password\": \"{password}\"}}' -k"
            cmd += f" > {_filename}.log"
            
            context.hosts[host].shellcmd._ssh.command(cmd, timeout=int(wait_time)).lower()
            for _ in range(int(wait_time) + 4):
                ret = context.hosts[host].shellcmd._ssh.command(f'cat {_filename}.log')
                if any(map(lambda s: s in ret, ret_mark)) or 'Invalid Authentication Information' in ret:
                    break
                time.sleep(1)

        elif row['console_type'] == 'Telnet':
            ret_mark = ['Firewall/VPN Router', 'MOXA']
            wait_time = int(wait_time) + 5
            cmd = f'''/bin/bash -c '(sleep 2; echo "{username}"; sleep 2; echo "{password}"; sleep {wait_time}) | \
            telnet {target_ip} > login_check_telnet.log' '''

            container_flag = True
            context.hosts[host].dut_ui.cli_cmd(cmd, detach=True)
            for _ in range(int(wait_time) + 5):
                ret = context.hosts[host].dut_ui.cli_cmd("/bin/bash -c 'cat login_check_telnet.log'")
                if any(map(lambda s: s in ret, ret_mark)) or 'Login incorrect' in ret:
                    break
                time.sleep(1)

        elif row['console_type'] == 'SSH':
            ret_mark = ['Firewall/VPN Router', 'MOXA']
            sleep_time = '3' if success_action_flag else '10'
            cmd = f'''/bin/bash -c 'sshpass -p {password} ssh -o "StrictHostKeyChecking no" \
            {username}@{target_ip} > login_check_ssh.log' '''

            container_flag = True
            context.hosts[host].dut_ui.cli_cmd(cmd, detach=True)
            for _ in range(int(wait_time) + 5):
                ret = context.hosts[host].dut_ui.cli_cmd("/bin/bash -c 'cat login_check_ssh.log'")
                if any(map(lambda s: s in ret, ret_mark)) or 'Permission denied' in ret:
                    break
                time.sleep(1)
            
        else:
            raise ValueError(f'This console type not supported: {console_type}')
        
        result_map[row['console_type']] = True if any(map(lambda s: s in ret, ret_mark)) else False
        print(result_map)
        time.sleep(2)

    if container_flag:
        context.hosts[host].dut_ui.cli_destroy()
    else :
        context.hosts[host].shellcmd._ssh.command('rm login_check_*')
    print(f'\n*---------End of login test--------*\n{result_map}')

    for key, value in result_map.items():
        if success_action_flag and not value:
            context.dut[device].go_config().set_authentication_login_mode(auth_mode='local')
            raise ValueError(f'Login should succeed but FAILS. Login test result: {result_map} \
            \n!!! RECOVERY: Restore DUT settings to the <Local> authentication mode. !!!')
        elif not success_action_flag and value: 
            context.dut[device].go_config().set_authentication_login_mode(auth_mode='local')
            raise ValueError(f'Login should fails but SUCCEED. Login test result: {result_map} \
            \n!!! RECOVERY: Restore DUT settings to the <Local> authentication mode. !!!')
